import numpy as np
import pytest

import splinepy


def test_BSplines(np_rng):
    """
    Perform checks on BSpline class
    """
    # No assertion query
    splinepy.BSpline(
        degrees=[1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 0.4, 0.5, 0.7, 1, 1]],
        control_points=np_rng.random((10, 2)),
    )
    # Check control points
    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid number of control points. "
        r"2 expected, but 9 were given.",
    ):
        splinepy.BSpline(
            degrees=[1],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((9, 3)),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Zero dimensional data points "
        r"can not constitute a spline.",
    ):
        splinepy.BSpline(
            degrees=[1],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((2, 0)),
        )

    with pytest.raises(
        ValueError,
        match=r"len\(knot_vectors\) \(2\) should match len\(self.degrees\) \(1\).",
    ):
        splinepy.BSpline(
            degrees=[1],
            knot_vectors=[[0, 0, 0.6, 0.3, 1, 1], [0, 0, 1, 1]],
            control_points=np.ones((8, 2)),
        )

    # Check degrees
    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid degree, degrees need to be positive. "
        r"Detected degree -4 along parametric dimension: 0",
    ):
        splinepy.BSpline(
            degrees=[-4],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((1, 1)),
        )


def test_NURBS(np_rng):
    """
    Perform checks on Nurbs class
    """
    # No assertion querz
    splinepy.NURBS(
        degrees=[1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 0.4, 0.5, 0.7, 1, 1]],
        control_points=np_rng.random((10, 2)),
        weights=np.ones(10),
    )
    # Check control points
    with pytest.raises(
        ValueError,
        match=r"len\(weights\) \(2\) should match len\(control_points\) \(9\).",
    ):
        splinepy.NURBS(
            degrees=[1],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((9, 3)),
            weights=np.ones(2),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Zero dimensional data points "
        r"can not constitute a spline.",
    ):
        splinepy.NURBS(
            degrees=[1],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((2, 0)),
            weights=np.ones(2),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - 0.6 0.3 Knots of parametric dimension \( 0 \) "
        r"are not in increasing order.",
    ):
        splinepy.NURBS(
            degrees=[1, 1],
            knot_vectors=[[0, 0, 0.6, 0.3, 1, 1], [0, 0, 1, 1]],
            control_points=np.ones((8, 2)),
            weights=np.ones(8),
        )

    # Check degrees
    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid degree, degrees need to be positive. "
        r"Detected degree -4 along parametric dimension: 0",
    ):
        splinepy.NURBS(
            degrees=[-4],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((1, 1)),
            weights=np.ones(1),
        )

    # Check weights
    with pytest.raises(
        ValueError,
        match=r"len\(weights\) \(19\) should match "
        r"len\(control_points\) \(2\).",
    ):
        splinepy.NURBS(
            degrees=[1],
            knot_vectors=[[0, 0, 1, 1]],
            control_points=np.ones((2, 1)),
            weights=np.ones(19),
        )


def test_Bezier():
    """Feed Bezier constructor with random data and see if the errors are
    meaningful"""
    # First create a reasonable spline
    splinepy.Bezier(degrees=[1, 1], control_points=np.ones((4, 1)))

    # Check assertions
    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid degree, degrees need to be positive. "
        r"Detected degree -4 along parametric dimension: 0",
    ):
        splinepy.Bezier(
            degrees=[-4],
            control_points=np.ones((1, 1)),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid number of control points. "
        r"9 expected, but 4 were given.",
    ):
        splinepy.Bezier(
            degrees=[2, 2],
            control_points=np.ones((4, 2)),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid number of control points. "
        r"24 expected, but 9 were given.",
    ):
        splinepy.Bezier(
            degrees=[2, 3, 1],
            control_points=np.ones((9, 3)),
        )


def test_RationalBezier(np_rng):
    """Feed Bezier constructor with random data and see if the errors are
    meaningful"""
    # First create a reasonable spline
    splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=np.ones((4, 1)),
        weights=np_rng.random(4),
    )

    # Check assertions
    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid degree, degrees need to be positive. "
        r"Detected degree -4 along parametric dimension: 0",
    ):
        splinepy.RationalBezier(
            degrees=[-4],
            control_points=np.ones((2, 4)),
            weights=np_rng.random(2),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid number of control points. "
        r"9 expected, but 4 were given.",
    ):
        splinepy.RationalBezier(
            degrees=[2, 2],
            control_points=np.ones((4, 1)),
            weights=np_rng.random(4),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Invalid number of control points. "
        r"24 expected, but 9 were given.",
    ):
        splinepy.RationalBezier(
            degrees=[2, 3, 1],
            control_points=np.ones((9, 2)),
            weights=np_rng.random(9),
        )

    with pytest.raises(
        RuntimeError,
        match=r"SPLINEPY ERROR - Zero dimensional data "
        r"points can not constitute a spline.",
    ):
        splinepy.RationalBezier(
            degrees=[2, 3, 1],
            control_points=np.ones((24, 0)),
            weights=np_rng.random(24),
        )

    with pytest.raises(
        ValueError,
        match=r"len\(weights\) \(9\) should match "
        r"len\(control_points\) \(24\).",
    ):
        splinepy.RationalBezier(
            degrees=[2, 3, 1],
            control_points=np.ones((24, 2)),
            weights=np_rng.random(9),
        )



import numpy as _np

from splinepy import utils as _utils


def permute_parametric_axes(spline, permutation_list, inplace=True):
    """
    Permutates the parametric dimensions

    This function can be used, e.g., to  interchange the parametric
    dimensions xi and eta in order to have them in the right orientation
    (applications in boundary condition definition or mfem export)

    Parameters
    ----------
    spline: Spline
    permutation_list : list
        New order of parametric dimensions
    inplace: bool
        Default is True. If True, modifies spline inplace, else, returns
        a modified_spline.

    Returns
    -------
    modified_spline : type(spline)
        spline with reordered parametric dimensions. iff `inplace=True`.
    """
    # Data collector for new spline object
    dict_spline = {}

    # Sanity checks
    if not isinstance(permutation_list, list):
        raise ValueError("Permutation list incomprehensive")
    if not set(range(spline.para_dim)) == set(permutation_list):
        raise ValueError("Permutation list invalid")

    _utils.log.debug("Permuting parametric axes...")

    # Update knot_vectors where applicable
    if "knot_vectors" in spline.required_properties:
        dict_spline["knot_vectors"] = [
            spline.knot_vectors[permutation_list[i]]
            for i in range(spline.para_dim)
        ]
    # Update degrees
    dict_spline["degrees"] = [
        spline.degrees[permutation_list[i]] for i in range(spline.para_dim)
    ]

    # Retrieve control mesh resolutions
    ctps_dims = spline.control_mesh_resolutions
    new_ctps_dims = [
        ctps_dims[permutation_list[i]] for i in range(spline.para_dim)
    ]
    n_ctps = spline.control_points.shape[0]

    # Map global to local index
    # i_glob = i + n_i * j  + n_i * n_j * k ...
    local_indices = _np.empty([n_ctps, spline.para_dim], dtype=int)
    global_indices = _np.arange(n_ctps, dtype=int)
    for i_p in range(spline.para_dim):
        local_indices[:, i_p] = global_indices % ctps_dims[i_p]
        global_indices -= local_indices[:, i_p]
        global_indices = _np.floor_divide(global_indices, ctps_dims[i_p])

    # Reorder indices
    local_indices[:] = local_indices[:, permutation_list]

    # Rearange global to local
    global_indices = _np.matmul(
        local_indices, _np.cumprod([1] + new_ctps_dims)[0:-1]
    )
    # Get inverse mapping
    global_indices = _np.argsort(global_indices)
    if "weights" in spline.required_properties:
        dict_spline["weights"] = spline.weights[global_indices]
    dict_spline["control_points"] = spline.control_points[global_indices, :]

    if inplace:
        _utils.log.debug("  applying permutation inplace")
        spline._new_core(**dict_spline)

        return None

    else:
        _utils.log.debug("  returning permuted spline")
        return type(spline)(**dict_spline)


def flip_axes(spline, axes, inplace=False):
    """
    Flip the parametric axes

    This function flips parametric axes without changing the parametric bounds
    of the spline. This can be used, e.g., to achieve conformity on a
    multipatch interface

    Parameters
    ----------
    spline: Spline
    axes : list
        List of axes to be flipped
    inplace: bool
        Default is True. If True, modifies spline inplace, else, returns
        a modified_spline.

    Returns
    -------
    modified_spline : type(spline)
        spline with reordered parametric dimensions. iff `inplace=True`.
    """

    # Sanity checks
    if isinstance(axes, int):
        axes = [axes]
    if not isinstance(axes, (list, tuple)):
        raise TypeError("Permutation list should be a list")

    # Create copy if not inplace
    if not inplace:
        spline = spline.copy()

    # Create a multi index object
    mi = spline.multi_index
    indexing = [slice(None, None, None)] * spline.para_dim

    # Loop over all axis to be flipped and prepare the flip
    for axis in axes:
        # Reverse control point indexing
        indexing[axis] = slice(None, None, -1)

        # Reverse knot vectors (keep parametric bounds unchanged)
        if spline.has_knot_vectors:
            para_bounds = spline.parametric_bounds
            spline.knot_vectors[axis][:] = list(
                para_bounds[0, axis]
                + para_bounds[1, axis]
                - _np.flip(spline.knot_vectors[axis])
            )
    # Convert list to tuple for indexing
    indexing = tuple(indexing)

    # Flip control points
    spline.cps[:] = spline.cps[mi[indexing]]

    # Do same with weights if applicable
    if spline.is_rational:
        spline.weights[:] = spline.weights[mi[indexing]]

    return spline



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class CubeVoid(_TileBase):
    """Void in form of an cuboid set into a unit cell.

    Parametrization acts on the cuboid's orientation as well as on its
    expansion (radii in x and y/z direction).

    Used in ADAMM test case for Hutchinson.

    See create_tile for more information

    .. raw:: html

        <p><a href="../_static/CubeVoid.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/CubeVoid.html" />

    """  # noqa: E501

    _dim = 3
    _para_dim = 3
    _evaluation_points = _np.array([[0.5, 0.5, 0.5]])
    _n_info_per_eval_point = 4

    # Aux values
    _sphere_ctps = _np.array(
        [
            [-0.5, -0.5, -0.5],
            [0.5, -0.5, -0.5],
            [-0.5, 0.5, -0.5],
            [0.5, 0.5, -0.5],
            [-0.5, -0.5, 0.5],
            [0.5, -0.5, 0.5],
            [-0.5, 0.5, 0.5],
            [0.5, 0.5, 0.5],
        ]
    )

    def _rotation_matrix_x(self, angle):
        cc, ss = _np.cos(angle), _np.sin(angle)
        return _np.array([[1, 0, 0], [0, cc, ss], [0, -ss, cc]])

    def _rotation_matrix_x_deriv(self, angle):
        cc, ss = _np.cos(angle), _np.sin(angle)
        return _np.array([[0, 0, 0], [0, -ss, cc], [0, -cc, -ss]])

    def _rotation_matrix_y(self, angle):
        cc, ss = _np.cos(angle), _np.sin(angle)
        return _np.array([[cc, 0, -ss], [0, 1, 0], [ss, 0, cc]])

    def _rotation_matrix_y_deriv(self, angle):
        cc, ss = _np.cos(angle), _np.sin(angle)
        return _np.array([[-ss, 0, -cc], [0, 1, 0], [cc, 0, -ss]])

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a full cuboid with a cuboidal void in the center, that has
        a rotation applied in the form

        T = RotY * RotX * Tz

        Parameters are stored as a 4 dimensional field. The entries represent:
        p_0 -> expansion along x-axis
        p_1 -> expansion along y/z axis
        p_2 -> rotation around x-axis
        p_3 -> rotation around y-axis

        Parameters
        ----------
        parameters : np.array
          Defines 4 values necessary for tile definition
          1. expansion in cuboid void in x-axis
          2. expansion of cuboid void in y- and z-axis
          3. rotation around x-axis
          4. rotation around y- and z-axis
        parameter_sensitivities: np.array
          Sensitivities of all 4 values with respect to some number of external
          design variables

        Returns
        -------
        microtile_list : list(splines)
        derivatives : list<list<splines>> / None
        """  # set to default if nothing is given
        if parameters is None:
            parameters = _np.array([0.5, 0.5, 0, 0]).reshape(1, 1, 4)

        # Create center ellipsoid
        # RotY * RotX * DIAG(r_x, r_yz) * T_base
        [r_x, r_yz, rot_x, rot_y] = parameters.flatten()

        # Check if user requests derivative splines
        if self.check_param_derivatives(parameter_sensitivities):
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        # Prepare auxiliary matrices and values
        diag = _np.diag([r_x, r_yz, r_yz])
        rotMx = self._rotation_matrix_x(rot_x)
        rotMy = self._rotation_matrix_y(rot_y)

        # Start assembly
        splines = []
        for i_derivative in range(n_derivatives + 1):
            spline_list = []
            # Constant auxiliary values
            if i_derivative == 0:
                v_one_half = 0.5
                v_one = 1.0
                v_zero = 0.0
                ctps = _np.einsum(
                    "ij,jk,kl,ml->mi",
                    rotMy,
                    rotMx,
                    diag,
                    self._sphere_ctps.copy(),
                    optimize=True,
                )
            else:
                [dr_x, dr_yz, drot_x, drot_y] = parameter_sensitivities[
                    0, :, i_derivative - 1
                ].flatten()
                v_one_half = 0.0
                v_one = 0.0
                v_zero = 0.0
                ddiag = _np.diag([dr_x, dr_yz, dr_yz])
                drotMx = self._rotation_matrix_x_deriv(rot_x)
                drotMy = self._rotation_matrix_y_deriv(rot_y)
                ctps_mat = (
                    drot_y
                    * _np.einsum(
                        "ij,jk,kl->il",
                        drotMy,
                        rotMx,
                        diag,
                    )
                    + drot_x * _np.einsum("ij,jk,kl->il", rotMy, drotMx, diag)
                    + _np.einsum("ij,jk,kl->il", rotMy, rotMx, ddiag)
                )
                ctps = _np.einsum("ij,kj->ki", ctps_mat, self._sphere_ctps)

            ctps += [v_one_half, v_one_half, v_one_half]
            # Start the assembly
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        [v_zero, v_zero, v_zero],
                        [v_one, v_zero, v_zero],
                        [v_zero, v_one, v_zero],
                        [v_one, v_one, v_zero],
                        ctps[0, :],
                        ctps[1, :],
                        ctps[2, :],
                        ctps[3, :],
                    ],
                )
            )
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        ctps[4, :],
                        ctps[5, :],
                        ctps[6, :],
                        ctps[7, :],
                        [v_zero, v_zero, v_one],
                        [v_one, v_zero, v_one],
                        [v_zero, v_one, v_one],
                        [v_one, v_one, v_one],
                    ],
                )
            )
            # Y-Axis
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        [v_zero, v_zero, v_zero],
                        [v_one, v_zero, v_zero],
                        ctps[0, :],
                        ctps[1, :],
                        [v_zero, v_zero, v_one],
                        [v_one, v_zero, v_one],
                        ctps[4, :],
                        ctps[5, :],
                    ],
                )
            )
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        ctps[2, :],
                        ctps[3, :],
                        [v_zero, v_one, v_zero],
                        [v_one, v_one, v_zero],
                        ctps[6, :],
                        ctps[7, :],
                        [v_zero, v_one, v_one],
                        [v_one, v_one, v_one],
                    ],
                )
            )
            # Z-Axis
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        [v_zero, v_zero, v_zero],
                        ctps[0, :],
                        [v_zero, v_one, v_zero],
                        ctps[2, :],
                        [v_zero, v_zero, v_one],
                        ctps[4, :],
                        [v_zero, v_one, v_one],
                        ctps[6, :],
                    ],
                )
            )
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1],
                    control_points=[
                        ctps[1, :],
                        [v_one, v_zero, v_zero],
                        ctps[3, :],
                        [v_one, v_one, v_zero],
                        ctps[5, :],
                        [v_one, v_zero, v_one],
                        ctps[7, :],
                        [v_one, v_one, v_one],
                    ],
                )
            )

            # Pass to output
            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        # Return results
        return (splines, derivatives)



"""
Export spline plots as svg files
"""

import numpy as np

import splinepy as spp

np_rng = np.random.default_rng()

box = spp.helpme.create.box(10, 5).bspline
box.elevate_degrees([0, 1])
box.insert_knots(0, [0.5])

# Create a simple curve
spline_curve = spp.Bezier(degrees=[2], control_points=[[0, 0], [2, 0], [2, 2]])
spp.io.svg.export("spline_curve_1.svg", spline_curve)

# Make it more complex
rational_spline_curve = spline_curve.nurbs
rational_spline_curve.elevate_degrees([0, 0])
rational_spline_curve.insert_knots(0, [0.4, 0.6])
rational_spline_curve.weights = np_rng.random(
    rational_spline_curve.cps.shape[0]
)
spp.io.svg.export("spline_curve_3.svg", rational_spline_curve)

# Export some surfaces
circle = spp.helpme.create.surface_circle(1)
spp.io.svg.export("spline_circle_1.svg", circle)

#
circle = circle.nurbs
circle.insert_knots(0, [0.333, 0.666])
circle.insert_knots(1, [0.333, 0.666])
circle.spline_data["field"] = circle
# svg export also respects many of the show_option configurations
circle.show_options["arrow_data"] = "field"
circle.show_options["arrow_data_scale"] = 0.4
circle.show_options["arrow_data_on"] = spp.utils.data.cartesian_product(
    [np.linspace(0, 1, 10) for _ in range(2)]
)
spp.io.svg.export("spline_circle_2.svg", circle, box_margins=1.0)

# Create a spline with a scalar field
distorted_rectangle = spp.Bezier(
    degrees=[2, 2],
    control_points=[
        [0.0, 0.0],
        [1.0, 0.5],
        [2.0, 0.0],
        [0.5, 0.8],
        [1.2, 1.0],
        [2.3, 1.0],
        [0.0, 2.0],
        [1.0, 1.5],
        [2.0, 1.8],
    ],
).bspline
distorted_rectangle.insert_knots(0, [0.2, 0.4, 0.6, 0.8])
distorted_rectangle.insert_knots(1, [0.25, 0.5, 0.75])

# Extract a field (jacobian determinant spline projection)
jacobian_determinant_spline = distorted_rectangle.create.determinant_spline()
distorted_rectangle.spline_data["jac_dets"] = jacobian_determinant_spline
distorted_rectangle.show_options["data"] = "jac_dets"
spp.io.svg.export(
    "spline_rectangle_with_field.svg", distorted_rectangle, box_margins=1.0
)



"""splinepy/splinepy/_base.py

Base type of splinepy
"""

from splinepy.utils import log as _log


class SplinepyBase:
    __slots__ = ()

    def __init_subclass__(cls, *args, **kwargs):
        """
        Add logger shortcut.
        """
        super().__init_subclass__(*args, **kwargs)
        cls._logi = _log.prepend_log("<" + cls.__qualname__ + ">", _log.info)
        cls._logd = _log.prepend_log("<" + cls.__qualname__ + ">", _log.debug)
        cls._logw = _log.prepend_log(
            "<" + cls.__qualname__ + ">", _log.warning
        )
        cls._loge = _log.prepend_log("<" + cls.__qualname__ + ">", _log.error)
        return super().__new__(cls)



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class Snappy(_TileBase):
    """Snap-through tile consisting of a thin truss and a thick truss that
    collide into each other.

    .. raw:: html

        <p><a href="../_static/Snappy.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/Snappy.html" />

    """  # noqa: E501

    _dim = 2
    _para_dim = 2
    # dummy values - not used
    _evaluation_points = _np.array([[0.5, 0.5]])
    _n_info_per_eval_point = 1

    def _closing_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,  # TODO
        closure=None,
        contact_length=0.1,
        a=0.1,
        b=0.2,
        r=0.15,
        **kwargs,  # noqa ARG002
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters: np.ndarray
          currently unused
        parameter_sensitivities: np.ndarray
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij unused
        closure : str
          string specifying the closing dimensions (e.g., x_min)
        contact_length : float
          the length of the wall that contacts neighboring microstructures
        a : float
          height/ thickness of the thinner/upper beam
        b : float
          height/ thickness of the lower/thicker beam
        r : float
          'radius' of the cubic bezier

        Results
        -------
        spline_list : list
        derivatives: list<list<splines>> / None
        """
        if closure is None:
            raise ValueError("No closing direction given")

        self.check_params(parameters)

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        spline_list = []
        v_zero = 0.0
        v_one_half = 0.5
        v_one = 1.0
        cl_2 = contact_length * 0.5
        cl_2_inv = 1 - contact_length * 0.5
        a_inv = v_one - a

        if closure == "y_min":
            # set points:
            spline_1 = _np.array(
                [
                    [v_zero, v_zero],
                    [cl_2, v_zero],
                    [v_zero, b],
                    [cl_2, b],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_1)
            )
            spline_2 = _np.array(
                [
                    [cl_2_inv, v_zero],
                    [v_one, v_zero],
                    [cl_2_inv, b],
                    [v_one, b],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_2)
            )
            spline_3 = _np.array(
                [
                    [v_zero, a_inv],
                    [cl_2, a_inv],
                    [v_zero, v_one],
                    [cl_2, v_one],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_3)
            )
            spline_4 = _np.array(
                [
                    [cl_2_inv, a_inv],
                    [v_one, a_inv],
                    [cl_2_inv, v_one],
                    [v_one, v_one],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_4)
            )

            spline_5 = _np.array(
                [
                    [v_one_half - cl_2, v_zero],
                    [v_one_half + cl_2, v_zero],
                    [v_one_half - cl_2, v_one_half],
                    [v_one_half + cl_2, v_one_half],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_5)
            )

            spline_6 = _np.array(
                [
                    [v_one_half - cl_2, v_one_half],
                    [v_one_half + cl_2, v_one_half],
                    [v_one_half - cl_2, v_one_half + a],
                    [v_one_half + cl_2, v_one_half + a],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_6)
            )

            spline_7 = _np.array(
                [
                    [cl_2, v_zero],
                    [cl_2 + r, v_zero],
                    [v_one_half - cl_2 - r, v_zero],
                    [v_one_half - cl_2, v_zero],
                    [cl_2, b],
                    [cl_2 + r, b],
                    [v_one_half - cl_2 - r, v_one_half],
                    [v_one_half - cl_2, v_one_half],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_7)
            )

            spline_8 = _np.array(
                [
                    [cl_2, v_zero],
                    [cl_2 + r, v_zero],
                    [v_one_half - cl_2 - r, v_zero],
                    [v_one_half - cl_2, v_zero],
                    [cl_2, v_one_half],
                    [cl_2 + r, v_one_half],
                    [v_one_half - cl_2 - r, b],
                    [v_one_half - cl_2, b],
                ]
            ) + [v_one_half, v_zero]

            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_8)
            )

            spline_9 = _np.array(
                [
                    [cl_2, a_inv],
                    [cl_2 + r, a_inv],
                    [v_one_half - cl_2 - r, v_one_half],
                    [v_one_half - cl_2, v_one_half],
                    [cl_2, v_one],
                    [cl_2 + r, v_one],
                    [v_one_half - cl_2 - r, v_one_half + a],
                    [v_one_half - cl_2, v_one_half + a],
                ]
            )

            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_9)
            )

            spline_10 = _np.array(
                [
                    [cl_2, v_one_half],
                    [cl_2 + r, v_one_half],
                    [v_one_half - cl_2 - r, a_inv],
                    [v_one_half - cl_2, a_inv],
                    [cl_2, v_one_half + a],
                    [cl_2 + r, v_one_half + a],
                    [v_one_half - cl_2 - r, v_one],
                    [v_one_half - cl_2, v_one],
                ]
            ) + [v_one_half, v_zero]

            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_10)
            )
            return spline_list
        elif closure == "y_max":
            spline_1 = _np.array(
                [
                    [v_zero, v_zero],
                    [cl_2, v_zero],
                    [v_zero, v_one],
                    [cl_2, v_one],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_1)
            )
            spline_2 = _np.array(
                [
                    [cl_2_inv, v_zero],
                    [v_one, v_zero],
                    [cl_2_inv, v_one],
                    [v_one, v_one],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_2)
            )
            spline_3 = _np.array(
                [
                    [v_one_half - cl_2, v_one_half - b],
                    [v_one_half + cl_2, v_one_half - b],
                    [v_one_half - cl_2, v_one],
                    [v_one_half + cl_2, v_one],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1], control_points=spline_3)
            )
            spline_4 = _np.array(
                [
                    [cl_2, v_zero],
                    [cl_2 + r, v_zero],
                    [v_one_half - cl_2 - r, v_one_half - b],
                    [v_one_half - cl_2, v_one_half - b],
                    [cl_2, v_one],
                    [cl_2 + r, v_one],
                    [v_one_half - cl_2 - r, v_one],
                    [v_one_half - cl_2, v_one],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_4)
            )
            spline_5 = _np.array(
                [
                    [cl_2, v_one_half - b],
                    [cl_2 + r, v_one_half - b],
                    [v_one_half - cl_2 - r, v_zero],
                    [v_one_half - cl_2, v_zero],
                    [cl_2, v_one],
                    [cl_2 + r, v_one],
                    [v_one_half - cl_2 - r, v_one],
                    [v_one_half - cl_2, v_one],
                ]
            ) + [v_one_half, v_zero]

            spline_list.append(
                _Bezier(degrees=[3, 1], control_points=spline_5)
            )
            return (spline_list, None)
        else:
            raise ValueError(
                "Closing tile is only implemented for y-enclosure"
            )

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,  # TODO
        contact_length=0.1,
        a=0.1,
        b=0.2,
        c=0.3,
        r=0.15,
        closure=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the wall
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array
          Currently, no parameter is used, (First test)
        parameter_sensitivities: list(np.ndarray)
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij, currently unused
        contact_length : float
          the length of the wall that contacts neighboring microstructures
        a : float
          height/ thickness of the thinner/upper beam
        b : float
          height/ thickness of the lower/thicker beam
        c : float
          offset to the upper beam (for consistent snap-through must fulfill
          2*c<1-b)
        r : float
          'radius' of the cubic bezier
        closure : str
          string specifying the closing dimensions (e.g., x_min)

        Returns
        -------
        microtile_list : list(splines)
        derivatives: list<list<splines>> / None
        """

        for param in [a, b, c, r, contact_length]:
            if not isinstance(param, float):
                raise ValueError(f"Invalid Type, {param} is not float")
            if param < 0:
                raise ValueError("Invalid parameter, must be > 0.")

        if not ((contact_length > 0) and (contact_length < 0.49)):
            raise ValueError("The length of a side must be in (0.01, 0.49)")

        # Check horizontal parameters
        if not ((r + contact_length) < 0.5):
            raise ValueError(
                "Inconsistent parameters, must fulfill : 2*r + contact_length"
                " < 0.5"
            )

        # Check vertical parameters
        if not ((2 * c + b) < 1.0) or a > c:
            raise ValueError(
                "Inconsistent parameters, must be 2*c<1-c and a<c"
            )

        if parameters is not None:
            raise NotImplementedError(
                "Parametriazation is not implemented for this tile yet"
            )

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        if closure is not None:
            return self._closing_tile(
                parameters=None,
                parameter_sensitivities=parameter_sensitivities,
                closure=closure,
                contact_length=contact_length,
                a=a,
                b=b,
                r=r,
                **kwargs,
            )

        v_zero = 0.0
        v_one_half = 0.5
        v_one = 1.0
        cl_2 = contact_length * 0.5
        cl_2_inv = 1 - contact_length * 0.5
        a_inv = v_one - a

        spline_list = []

        # set points:
        spline_1 = _np.array(
            [
                [v_zero, v_zero],
                [cl_2, v_zero],
                [v_zero, b],
                [cl_2, b],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_1))
        spline_2 = _np.array(
            [
                [cl_2_inv, v_zero],
                [v_one, v_zero],
                [cl_2_inv, b],
                [v_one, b],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_2))
        spline_3 = _np.array(
            [
                [v_zero, a_inv],
                [cl_2, a_inv],
                [v_zero, v_one],
                [cl_2, v_one],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_3))
        spline_4 = _np.array(
            [
                [cl_2_inv, a_inv],
                [v_one, a_inv],
                [cl_2_inv, v_one],
                [v_one, v_one],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_4))

        spline_5 = _np.array(
            [
                [v_one_half - cl_2, v_one_half - b],
                [v_one_half + cl_2, v_one_half - b],
                [v_one_half - cl_2, v_one_half],
                [v_one_half + cl_2, v_one_half],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_5))

        spline_6 = _np.array(
            [
                [v_one_half - cl_2, v_one_half],
                [v_one_half + cl_2, v_one_half],
                [v_one_half - cl_2, v_one_half + a],
                [v_one_half + cl_2, v_one_half + a],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=spline_6))

        spline_7 = _np.array(
            [
                [cl_2, v_zero],
                [cl_2 + r, v_zero],
                [v_one_half - cl_2 - r, v_one_half - b],
                [v_one_half - cl_2, v_one_half - b],
                [cl_2, b],
                [cl_2 + r, b],
                [v_one_half - cl_2 - r, v_one_half],
                [v_one_half - cl_2, v_one_half],
            ]
        )

        spline_list.append(_Bezier(degrees=[3, 1], control_points=spline_7))

        spline_8 = _np.array(
            [
                [cl_2, v_one_half - b],
                [cl_2 + r, v_one_half - b],
                [v_one_half - cl_2 - r, v_zero],
                [v_one_half - cl_2, v_zero],
                [cl_2, v_one_half],
                [cl_2 + r, v_one_half],
                [v_one_half - cl_2 - r, b],
                [v_one_half - cl_2, b],
            ]
        ) + [v_one_half, v_zero]

        spline_list.append(_Bezier(degrees=[3, 1], control_points=spline_8))

        spline_9 = _np.array(
            [
                [cl_2, a_inv],
                [cl_2 + r, a_inv],
                [v_one_half - cl_2 - r, v_one_half],
                [v_one_half - cl_2, v_one_half],
                [cl_2, v_one],
                [cl_2 + r, v_one],
                [v_one_half - cl_2 - r, v_one_half + a],
                [v_one_half - cl_2, v_one_half + a],
            ]
        )

        spline_list.append(_Bezier(degrees=[3, 1], control_points=spline_9))

        spline_10 = _np.array(
            [
                [cl_2, v_one_half],
                [cl_2 + r, v_one_half],
                [v_one_half - cl_2 - r, a_inv],
                [v_one_half - cl_2, a_inv],
                [cl_2, v_one_half + a],
                [cl_2 + r, v_one_half + a],
                [v_one_half - cl_2 - r, v_one],
                [v_one_half - cl_2, v_one],
            ]
        ) + [v_one_half, v_zero]

        spline_list.append(_Bezier(degrees=[3, 1], control_points=spline_10))

        return (spline_list, None)



"""
Helps you map derivatives of fields and basis function into the physical
(image) domain
"""

import numpy as _np

from splinepy import utils as _utils
from splinepy._base import SplinepyBase as _SplinepyBase


class Mapper(_SplinepyBase):
    r"""Map expressions and derivatives into the physical domain using a
    geometric mapper

    The mapper has access to two fields: the solution field and the underlying
    geometry representation, both are assumed to be spline-functions. Hence, we
    assume a geometry representation in the form:

    .. math:: x_i (\mathbf{u}) = \sum_b N^b (\mathbf{u}) x^b_i

    as well as a field representation in the form :

    .. math:: f_i (\mathbf{u}) = \sum_a N^a (\mathbf{u}) f^a_i

    With :math:`x_i`, :math:`f_i` being (vector valued) geometry and solution
    field functions, with their respective shape/basis functions :math:`N^b`,
    :math:`N^a` and their associated coefficients/control points, :math:`x^b_i`
    and :math:`f^a_i`. Further, using the definition of the Jacobian and it's
    inverse we can write

    .. math:: \mathbf{J}^{-1} = \bar{\mathbf{J}}
    .. math:: J_{ij} = \frac{\partial x_i}{\partial u_j}
    .. math:: \bar{J}_{ij} = \frac{\partial u_i}{\partial x_j}

    Basis function derivatives of the first order can be mapped into the
    physical domain, using the following equation:

    .. math:: \frac{\partial N^a}{\partial x_i} =
      \frac{ \partial N^a}{\partial u_j} \frac{ \partial u_j}{\partial x_i} =
      \frac{\partial N^a}{ \partial u_j} \bar{J}_{ji}

    Using the definition of the solution field, field gradients (and
    divergences) can be computed in a similar fashion

    .. math:: \frac{\partial f_i}{\partial x_j} =
        \sum_a \frac{\partial N^a}{\partial x_j} f^a_i

    Divergences are computed by summing up the respective components.

    Second order derivatives of the basis function values with respect to the
    reference coordinates are written as (similar for :math:`H^b`):

    .. math:: H^a_{ij} = \frac{\partial^2 N^a}{\partial u_i\partial u_j}

    The hessians can be mapped into physical space using the following
    equation:

    .. math:: \frac{\partial^2 N^a}{\partial x_i\partial x_j} =
      \bar{J}_{li} \left(
        H^a_{kl} - \frac{\partial N^a}{\partial u_n}\bar{J}_{nm} x^b_m H^b_{lk}
      \right) \bar{J}_{kj}

    Here, the expression :math:`x^b_mH^b_{lk}` represents the Hessian of the
    geometric reference spline :math:`\tilde{H}_{mlk}`, which can sometimes be
    calculated more efficiently using spline methods implemented in SplineLib
    and Bezman.

    Here we used the following identity:

    .. math:: \frac{\partial \bar{J}_{li} }{\partial u_k} =
      - \bar{J}_{lm} \frac{\partial J_{mn}}{\partial u_k} \bar{J}_{ni}\\


    Parameters
    ----------
    field : spline
      Field that is to be mapped
    reference : spline
      Reference for mapping (geometry)
    """

    def __init__(self, field, reference):
        self._field_reference = field
        self._geometry_reference = reference

        # Check the parametric dimensions
        if not reference.para_dim == field.para_dim:
            raise ValueError("Parametric dimension mismatch")
        if not reference.para_dim == reference.dim:
            raise ValueError(
                "Mismatch between physical and parametric dimension for "
                "geometry representation"
            )
        # Easy access
        self._para_dim = reference.para_dim

    def basis_function_derivatives(
        self,
        queries,
        gradient=False,
        hessian=False,
        laplacian=False,
        nthreads=None,
    ):
        """Function to retrieve more than one basis function derivative

        More efficient implementation if more than one derivative is required,
        as many values can be precalculated. See class documentation for more
        details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Evaluation points in the parametric domain
        gradient : bool
          Evaluate Basis Function Gradient mapped into the physical domain
        hessian : bool
          Evaluate Basis Function Hessian mapped into the physical domain
        laplacian : bool
          Evaluate Basis Function Laplacian mapped into the physical domain
        nthreads: int
          Number of threads available for the computation

        Returns
        -------
        results : dict
          Dictionary with required values stored with same name as function
          arguments
        """
        # Naming convention for indices (more indices follow from
        # implementation details in class documentation):
        # q : query-point
        # s : support-index
        # a : solution field shape_function
        # b : geometry field shape_function
        self._logd("Evaluating basis function gradients in physical space")
        queries = _utils.data.enforce_contiguous(queries, dtype="float64")

        # Precompute required values
        invjacs = _np.linalg.inv(
            self._geometry_reference.jacobian(queries, nthreads)
        )
        results = {}

        bf_gradients = _np.empty(
            (
                queries.shape[0],
                _np.prod(self._field_reference.degrees + 1),
                self._para_dim,
            )
        )
        (
            bf_gradients[:, :, 0],
            support,
        ) = self._field_reference.basis_derivative_and_support(
            queries=queries,
            # Array size M with [0, ..., 0, 1, 0, ..., 0] with 1 at position k
            orders=_np.eye(1, M=self._para_dim, k=0),
            nthreads=nthreads,
        )
        results["support"] = support
        for i in range(1, self._para_dim):
            bf_gradients[:, :, i] = self._field_reference.basis_derivative(
                queries=queries,
                orders=_np.eye(1, M=self._para_dim, k=i),
                nthreads=nthreads,
            )
        if gradient:
            results["gradient"] = _np.einsum(
                "qsi,qij->qsj", bf_gradients, invjacs, optimize=True
            )
            results["support"] = support

        if hessian or laplacian:
            # Retrieve basis function hessians from both the geometry as from
            # the field
            bf_hessians = _np.empty(
                (
                    queries.shape[0],
                    _np.prod(self._field_reference.degrees + 1),
                    self._para_dim,
                    self._para_dim,
                )
            )
            for i in range(self._para_dim):
                for j in range(i, self._para_dim):
                    bf_hessians[:, :, i, j] = (
                        self._field_reference.basis_derivative(
                            queries=queries,
                            orders=_np.eye(1, M=self._para_dim, k=i)
                            + _np.eye(1, M=self._para_dim, k=j),
                            nthreads=nthreads,
                        )
                    )
                    if i != j:
                        bf_hessians[:, :, j, i] = bf_hessians[:, :, i, j]
            # This is unnecessary if isoparametric (but if only field is high
            # order, this is more efficient)
            geo_hessians = _np.empty(
                (
                    queries.shape[0],
                    self._para_dim,  # dim==para_dim
                    self._para_dim,
                    self._para_dim,
                )
            )
            for i in range(self._para_dim):
                for j in range(i, self._geometry_reference.para_dim):
                    geo_hessians[:, :, i, j] = (
                        self._geometry_reference.derivative(
                            queries=queries,
                            orders=_np.eye(1, M=self._para_dim, k=i)
                            + _np.eye(1, M=self._para_dim, k=j),
                            nthreads=nthreads,
                        )
                    )
                    if i != j:
                        geo_hessians[:, :, j, i] = geo_hessians[:, :, i, j]

            # Overwrite bf_hessians (see documentation for indices)
            bf_hessians -= _np.einsum(
                "qan,qnm,qmlk->qalk",
                bf_gradients,
                invjacs,
                geo_hessians,
                optimize=True,
            )
        if hessian:
            results["hessian"] = _np.einsum(
                "eli,ealk,ekj->eaij",
                invjacs,
                bf_hessians,
                invjacs,
                optimize=True,
            )
            results["support"] = support
        if laplacian:
            if hessian:
                results["laplacian"] = _np.einsum(
                    "eaii->ea", results["hessian"], optimize=True
                )
            else:
                results["laplacian"] = _np.einsum(
                    "eli,ealk,eki->ea",
                    invjacs,
                    bf_hessians,
                    invjacs,
                    optimize=True,
                )
                results["support"] = support

        return results

    def field_derivatives(
        self,
        queries,
        gradient=False,
        divergence=False,
        hessian=False,
        laplacian=False,
        basis_function_values=False,
        nthreads=None,
    ):
        """Function to retrieve more than one field derivative

        More efficient implementation if more than one derivative is required.
        Can also return basis function values if both are required, e.g., for
        some assembly. See class documentation for more details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Evaluation points in the parametric domain
        gradient : bool
          Evaluate Gradient mapped into the physical domain
        divergence : bool
          Evaluate Divergenec of a vector field
        hessian : bool
          Evaluate Hessian mapped into the physical domain
        laplacian : bool
          Evaluate Laplacian mapped into the physical domain
        basis_function_values : bool
          Return basis function derivatives in dictionary
        nthreads: int
          Number of threads available for the computation

        Returns
        -------
        results : dict
          Dictionary with required values stored with same name as function
          arguments (basis function derivatives as dictionary in dictionary)
        """
        # Remark : In terms of computations, it might actually be more
        # efficient, to calculate the derivatives of the spline values
        # directly, rather than to precompute al basis function derivatives, as
        # more efficient methods might be implemented in both SplineLib and
        # Bezman, however, for sake of simplicity, we will stick to the basis
        # function based implementation here, as basis function derivatives are
        # mostly required in IGA applications.

        # Compute required basis function values
        as_dictionary = self.basis_function_derivatives(
            queries=queries,
            gradient=(gradient or divergence),
            hessian=hessian,
            laplacian=laplacian,
            nthreads=nthreads,
        )

        results = {}
        supports = as_dictionary["support"]
        if basis_function_values:
            results["basis_function_values"] = as_dictionary
        # Start computation
        if gradient:
            results["gradient"] = _np.einsum(
                "qsd,qsv->qvd",
                as_dictionary["gradient"],
                self._field_reference.control_points[supports, :],
                optimize=True,
            )
        if divergence:
            # Can only be performed on vector fields with para_dim=dim
            if self._field_reference.para_dim == self._field_reference.dim:
                if gradient:
                    results["divergence"] = _np.einsum(
                        "qii->q", results["gradient"], optimize=True
                    )
                else:
                    results["divergence"] = _np.einsum(
                        "qsd,qsd->q",
                        as_dictionary["gradient"],
                        self._field_reference.control_points[supports, :],
                        optimize=True,
                    )
            else:
                raise ValueError(
                    "Divergence can only be performed on vector fields with "
                    "para_dim = dim"
                )

        if hessian:
            results["hessian"] = _np.einsum(
                "qsij,qsv->qvij",
                as_dictionary["hessian"],
                self._field_reference.control_points[supports, :],
                optimize=True,
            )
        if laplacian:
            if hessian:
                results["laplacian"] = _np.einsum(
                    "qdii->qd", results["hessian"], optimize=True
                )
            else:
                results["laplacian"] = _np.einsum(
                    "qs,qsd->qd",
                    as_dictionary["laplacian"],
                    self._field_reference.control_points[supports, :],
                    optimize=True,
                )

        return results

    def basis_gradient_and_support(self, queries, nthreads=None):
        """Map gradient of basis functions into the physical domain

        Calls self.basis_function_derivatives internally, see class
        documentation for more details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
        nthreads: int

        Returns
        --------
        gradient: (n, prod(degrees + 1), para_dim) np.ndarray
        support: (n, prod(degrees + 1)) np.ndarray
        """
        as_dict = self.basis_function_derivatives(
            queries,
            gradient=True,
            hessian=False,
            laplacian=False,
            nthreads=nthreads,
        )
        return (as_dict["gradient"], as_dict["support"])

    def gradient(self, queries, nthreads=None):
        r"""Map gradient field into the physical domain

        Gradient is in form

        .. math:: J^k_i = \frac{\partial f^k}{\partial x_i}

        See class documentation for more details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Points for evaluation
        nthreads: int
          Threads used for calculation

        Returns
        --------
        gradient: (n, dim, para_dim) np.ndarray
        """
        return self.field_derivatives(
            queries,
            gradient=True,
            divergence=False,
            hessian=False,
            laplacian=False,
            basis_function_values=False,
            nthreads=nthreads,
        )["gradient"]

    def divergence(self, queries, nthreads=None):
        """Map field divergence (where applicable) into the physical domain

        Returns scalar values for each query. See class documentation for more
        details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Points for evaluation
        nthreads: int
          Threads used for calculation

        Returns
        --------
        gradient: (n) np.ndarray
        """
        if self._field_reference.para_dim != self._field_reference.dim:
            raise ValueError(
                "Divergence can only be performed on vector fields with "
                "para_dim = dim"
            )
        return self.field_derivatives(
            queries,
            gradient=False,
            divergence=True,
            hessian=False,
            laplacian=False,
            basis_function_values=False,
            nthreads=nthreads,
        )["divergence"]

    def basis_hessian_and_support(self, queries, nthreads=None):
        """Map hessian of basis functions into the physical domain

        Calls self.basis_function_derivatives internally, see class
        documentation for more details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
        nthreads: int

        Returns
        --------
        hessians: (n, prod(degrees + 1), para_dim, para_dim) np.ndarray
        support: (n, prod(degrees + 1)) np.ndarray
        """
        as_dict = self.basis_function_derivatives(
            queries,
            gradient=False,
            hessian=True,
            laplacian=False,
            nthreads=nthreads,
        )
        return (as_dict["hessian"], as_dict["support"])

    def hessian(self, queries, nthreads=None):
        r"""Map hessian field into the physical domain

        Hessian is in form

        .. math:: H^k_{ij} = \frac{\partial^2 f_k}{\partial x_i \partial x_j}

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Points for evaluation
        nthreads: int
          Threads used for calculation

        Returns
        --------
        gradient: (n, prod(degrees + 1), paradim) np.ndarray
        """
        return self.field_derivatives(
            queries,
            gradient=False,
            divergence=False,
            hessian=True,
            laplacian=False,
            basis_function_values=False,
            nthreads=nthreads,
        )["hessian"]

    def basis_laplacian_and_support(self, queries, nthreads=None):
        """Map laplacian of basis functions into the physical domain

        Calls self.basis_function_derivatives internally, see class
        documentation for more details.

        Parameters
        ----------
        queries: (n, para_dim) array-like
        nthreads: int

        Returns
        --------
        laplacian: (n, prod(degrees + 1)) np.ndarray
        support: (n, prod(degrees + 1)) np.ndarray
        """
        as_dict = self.basis_function_derivatives(
            queries,
            gradient=False,
            hessian=False,
            laplacian=True,
            nthreads=nthreads,
        )
        return (as_dict["laplacian"], as_dict["support"])

    def laplacian(self, queries, nthreads=None):
        """Map laplacian field into the physical domain

        Parameters
        ----------
        queries: (n, para_dim) array-like
          Points for evaluation
        nthreads: int
          Threads used for calculation

        Returns
        --------
        laplacian: (n) np.ndarray
        """
        return self.field_derivatives(
            queries,
            gradient=False,
            divergence=False,
            hessian=False,
            laplacian=True,
            basis_function_values=False,
            nthreads=nthreads,
        )["laplacian"]



import numpy as np
import pytest

import splinepy

# frequently used fixtures
all_splinetypes = ("bspline_3p3d", "nurbs_3p3d", "bspline_2p2d", "nurbs_2p2d")


def test_extraction(np_rng):
    """
    test the extraction of beziers
    """
    # Define some splines
    b = splinepy.BSpline(
        knot_vectors=[[0, 0, 0, 1, 2, 2, 3, 4, 4, 4]],
        degrees=[2],
        control_points=np_rng.random((7, 2)),
    )
    n = splinepy.NURBS(
        knot_vectors=[[0, 0, 0, 0, 1, 1, 2, 3, 4, 4, 4, 4]],
        degrees=[3],
        control_points=np_rng.random((8, 2)),
        weights=np_rng.random((8, 1)),
    )

    # Extract Beziers
    b_beziers = b.extract_bezier_patches()
    n_beziers = n.extract_bezier_patches()

    # Loop over knot_spans and test at random points
    for offset in range(4):
        queries = np_rng.random((20, 1))
        assert np.allclose(
            b.evaluate(queries + offset),
            b_beziers[offset].evaluate(queries),
        )

        assert np.allclose(
            n.evaluate(queries + offset),
            n_beziers[offset].evaluate(queries),
        )


@pytest.mark.parametrize("splinetype", all_splinetypes)
def test_extraction_matrices(splinetype, np_rng, request):
    spline = request.getfixturevalue(splinetype)

    spline.elevate_degrees(list(range(spline.para_dim)))
    for i in range(spline.para_dim):
        spline.insert_knots(i, np_rng.random(3))

    n_matrices = spline.knot_insertion_matrix(beziers=True)
    beziers_n = spline.extract_bezier_patches()
    for m, b in zip(n_matrices, beziers_n):
        # Test matrices m against spline ctps
        if "nurbs" in splinetype:
            assert np.allclose(b.weights, m @ spline.weights)
            assert np.allclose(
                b.control_points,
                (m @ (spline.control_points * spline.weights)) / b.weights,
            )
        elif "bspline" in splinetype:
            assert np.allclose(b.control_points, m @ spline.control_points)
        else:
            pytest.fail()


def test_extraction_0th_degree():
    """Edge case for 0th degree"""
    test = splinepy.BSpline(
        degrees=[0, 0],
        control_points=[
            [0],
            [1],
            [2],
            [3],
        ],
        knot_vectors=[[0, 0.5, 1], [0, 0.5, 1]],
    )
    for i, patch in enumerate(test.extract.beziers()):
        assert np.allclose(test.cps[i], patch.cps[0])



from splinepy.io import cats, gismo, iges, ioutils, irit, json, mfem, npz, svg
from splinepy.io.default import export, load

__all__ = [
    "cats",
    "export",
    "gismo",
    "iges",
    "ioutils",
    "irit",
    "json",
    "load",
    "mfem",
    "npz",
    "svg",
]



import numpy as np
import pytest

import splinepy

# fixtures used
all_3p3d_splines = (
    "rational_bezier_3p3d",
    "bezier_3p3d",
    "bspline_3p3d",
    "nurbs_3p3d",
)


@pytest.mark.parametrize("splinetype", all_3p3d_splines)
def test_permute_parametric_axes(request, splinetype, queries_3D):
    """
    test permute
    """
    # Define spline
    spline = request.getfixturevalue(splinetype)

    # define permutation
    permutation = [2, 0, 1]

    # return permuted
    # make more work
    spline.elevate_degrees(1)
    spline.elevate_degrees(2)
    spline.elevate_degrees(2)
    if "knot_vectors" in spline.required_properties:
        spline.insert_knots(0, [0.4, 0.7, 0.8])
        spline.insert_knots(1, [0.1, 0.2])
        spline.insert_knots(2, [0.3, 0.5, 0.6, 0.9])

    perm = splinepy.helpme.reparametrize.permute_parametric_axes(
        spline, permutation, inplace=False
    )
    queries = np.asarray(queries_3D)

    assert np.allclose(
        spline.evaluate(queries),
        perm.evaluate(queries[:, permutation]),
    ), f"{perm.whatami} failed to permute."

    # inplace
    perm = spline.copy()
    splinepy.helpme.reparametrize.permute_parametric_axes(
        perm, permutation, inplace=True
    )
    queries = np.asarray(queries_3D)

    assert np.allclose(
        spline.evaluate(queries),
        perm.evaluate(queries[:, permutation]),
    ), f"{perm.whatami} failed to permute inplace."


@pytest.mark.parametrize("splinetype", all_3p3d_splines)
def test_flip_axes(splinetype, request, queries_3D):
    """
    test invert axes
    """
    # Define spline
    spline = request.getfixturevalue(splinetype)

    flipped_axes = [1, 2]

    queries = np.asarray(queries_3D)
    flipped_queries = queries.copy()
    for axis in flipped_axes:
        # Queries are all in unit cube
        flipped_queries[:, axis] = 1 - flipped_queries[:, axis]

    # With copy
    new_spline = splinepy.helpme.reparametrize.flip_axes(
        spline, flipped_axes, inplace=False
    )

    assert np.allclose(
        spline.evaluate(queries),
        new_spline.evaluate(flipped_queries),
    ), f"{spline.whatami} failed to flip axis."

    # in place
    spline_copy = spline.copy()
    new_spline2 = splinepy.helpme.reparametrize.flip_axes(
        spline_copy, flipped_axes, inplace=True
    )

    # Check if truly in place
    assert (
        new_spline2 is spline_copy
    ), "Operation has not been performed in place"

    assert np.allclose(
        new_spline2.evaluate(queries),
        new_spline.evaluate(queries),
    ), f"{spline.whatami} failed to flip axis."



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import splinepy

# sys.path.insert(0, os.path.abspath("../../"))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "splinepy"
copyright = "2022, Jaewook Lee"  # noqa A001
author = "Jaewook Lee"
release = splinepy.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_mdinclude",
]


templates_path = ["_templates"]
exclude_patterns = []

pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_logo = "_static/splinepy_name.png"
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/tataratat/splinepy",
            "icon": "fa-brands fa-square-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/splinepy/",
            "icon": "fa-solid fa-box",
        },
    ],
    "navigation_with_keys": False,
}
html_favicon = "_static/favicon.png"
# html_favicon = "_static/favicon.ico"
html_static_path = ["_static"]
html_css_files = ["style.css"]

autodoc_default_options = {
    "autosummary": True,
}

autosummary_context = {
    "skipmethods": ["__init__"],
}



import argparse

import gustaf as gus
import numpy as np
import vedo

import splinepy

# this will cause a deprecated warning for newer version of vedo.
# for backward compatibility, we keep this one.
vedo.settings.screenshot_transparent_background = 1

if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        description=(
            "Creates a fanfare. Default option corresponds to "
            "the tataratat logo."
        )
    )
    arg_parser.add_argument(
        "--r_mouth",
        dest="r_mouth",
        default=0.2,
        help="Mouth radius for mouth piece.",
    )
    arg_parser.add_argument(
        "--r_shaft",
        dest="r_shaft",
        default=0.1,
        help="Shaft radius for mouth piece.",
    )
    arg_parser.add_argument(
        "--r_rounding",
        dest="r_rounding",
        default=0.05,
        help="Radius offset for mouth piece rounding.",
    )
    arg_parser.add_argument(
        "--l_mouth",
        dest="l_mouth",
        default=0.4,
        help="Length of the mouth piece.",
    )
    arg_parser.add_argument(
        "--l_cone",
        dest="l_cone",
        default=1.2,
        help="Length of the cone.",
    )
    arg_parser.add_argument(
        "--r_cone",
        dest="r_cone",
        default=0.8,
        help="Radius of the cone.",
    )
    arg_parser.add_argument(
        "--l_shaft",
        dest="l_shaft",
        default=4.0,
        help="Length of the shaft.",
    )
    arg_parser.add_argument(
        "--r_curve",
        dest="r_curve",
        default=0.3,
        help="Radius of connecting curves.",
    )
    args = arg_parser.parse_args()

    # Let's draw a fanfare
    fanfare = []

    # cast all args to float
    # Mouthpiece
    r_mouth = float(args.r_mouth)
    r_shaft = float(args.r_shaft)
    r_rounding = float(args.r_rounding)
    l_mouth = float(args.l_mouth)
    # Cone
    l_cone = float(args.l_cone)
    r_cone = float(args.r_cone)
    # Shaft
    l_shaft = float(args.l_shaft)
    curved_radius = float(args.r_curve)

    # Create fanfare part by part
    mouth_piece = splinepy.BSpline(
        knot_vectors=[[0, 0, 0, 0.25, 0.5, 0.5, 0.75, 1, 1, 1]],
        degrees=[2],
        control_points=[
            [l_mouth, r_mouth - r_rounding],
            [l_mouth + r_rounding, r_mouth - r_rounding],
            [l_mouth + r_rounding, r_mouth],
            [l_mouth, r_mouth],
            [0.5 * l_mouth, r_mouth],
            [0.5 * l_mouth, r_shaft],
            [0.0, r_shaft],
        ],
    ).nurbs.create.revolved(
        axis=[1, 0, 0],
        center=[0, 0, 0],
        angle=360,
        degree=True,
        n_knot_spans=4,
    )

    isqrt2 = 2 ** (-0.5)
    cross_section = splinepy.NURBS(
        degrees=[2],
        knot_vectors=[[0, 0, 0, 0.25, 0.25, 0.5, 0.5, 0.75, 0.75, 1, 1, 1]],
        control_points=(
            np.array(
                [
                    [0.0, 0, 0],
                    [0.0, 0.5, -0.5],
                    [0.0, 1, 0],
                    [0.0, 1 + 0.5, 0.5],
                    [0.0, 1, 1],
                    [0.0, 0.5, 1 + 0.5],
                    [0.0, 0, 1],
                    [0.0, -0.5, 0.5],
                    [0, 0, 0],
                ]
            )
            + np.array([0, -0.5, -0.5])
        )
        * (r_shaft * 2 * isqrt2),
        weights=[1, isqrt2, 1, isqrt2, 1, isqrt2, 1, isqrt2, 1],
    )

    shaft = cross_section.create.extruded(extrusion_vector=[-l_shaft, 0, 0])

    lower_shaft = shaft.copy()
    lower_shaft.control_points += [0, -2 * curved_radius, 0]
    upper_shaft = shaft.copy()
    upper_shaft.control_points += [
        0,
        -2 * curved_radius + 2 * isqrt2 * curved_radius,
        -2 * isqrt2 * curved_radius,
    ]

    cross_section.control_points += [-l_shaft, 0, 0]

    curved_piece = cross_section.nurbs.create.revolved(
        axis=[0.0, 0, 1], center=[-l_shaft, -curved_radius, 0], angle=180
    )
    cross_section.control_points += [l_shaft, -2 * curved_radius, 0]
    second_curved_piece = cross_section.nurbs.create.revolved(
        axis=[0, 1.0, 1.0],
        center=[
            0,
            -2 * curved_radius + isqrt2 * curved_radius,
            -isqrt2 * curved_radius,
        ],
        angle=180,
    )

    cone_part = splinepy.BSpline(
        degrees=[2],
        knot_vectors=[[0, 0, 0, 0.5, 1, 1, 1]],
        control_points=[
            [0.0, r_shaft, 0.0],
            [-0.25 * l_cone, r_shaft, 0.0],
            [-l_cone, 0.5 * (r_cone + r_shaft), 0.0],
            [-l_cone, r_cone, 0.0],
        ],
    ).nurbs.create.revolved(
        axis=[1, 0, 0],
        center=[0, 0, 0],
        angle=360,
        degree=True,
    )
    cone_part.control_points += [
        -l_shaft,
        -2 * curved_radius + 2 * isqrt2 * curved_radius,
        -2 * isqrt2 * curved_radius,
    ]

    fanfare.append(mouth_piece)
    fanfare.append(curved_piece)
    fanfare.append(second_curved_piece)
    fanfare.append(upper_shaft)
    fanfare.append(cone_part)
    fanfare.append(shaft)
    fanfare.append(lower_shaft)

    # show fanfare - prepare only surface and full spline components
    face_showables = []
    spline_showables = []
    for f in fanfare:
        showables = f.showable(resolutions=500, c=(255, 215, 0))
        faces = showables["spline"]
        face_showables.append(faces)
        spline_showables.extend(list(showables.values()))

    cam = {
        "pos": (-11.90193, 5.995478, 3.057389),
        "focalPoint": (-2.817062, -0.6220471, -0.3948362),
        "viewup": (0.3271027, 0.7493888, -0.5756911),
        "distance": 11.75773,
        "clippingRange": (6.030973, 19.21880),
    }

    gus.show(face_showables, spline_showables, cam=cam)



"""splinepy/microstructure/tiles/__init__.py.

Interface for tools and generators creating simple microstructures.
"""

# required in TileLib
from splinepy.microstructure.tiles import (
    armadillo,
    chi,
    cross_2d,
    cross_3d_linear,
    cube_void,
    double_lattice,
    ellips_v_oid,
    hollow_cube,
    hollow_octagon,
    hollow_octagon_extrude,
    inverse_cross_3d,
    snappy,
    tile_base,
)
from splinepy.microstructure.tiles.armadillo import Armadillo
from splinepy.microstructure.tiles.chi import Chi
from splinepy.microstructure.tiles.cross_2d import Cross2D
from splinepy.microstructure.tiles.cross_3d import Cross3D
from splinepy.microstructure.tiles.cross_3d_linear import Cross3DLinear
from splinepy.microstructure.tiles.cube_void import CubeVoid
from splinepy.microstructure.tiles.double_lattice import DoubleLattice
from splinepy.microstructure.tiles.ellips_v_oid import EllipsVoid
from splinepy.microstructure.tiles.hollow_cube import HollowCube
from splinepy.microstructure.tiles.hollow_octagon import HollowOctagon
from splinepy.microstructure.tiles.hollow_octagon_extrude import (
    HollowOctagonExtrude,
)
from splinepy.microstructure.tiles.inverse_cross_3d import InverseCross3D
from splinepy.microstructure.tiles.snappy import Snappy
from splinepy.microstructure.tiles.tile_base import TileBase
from splinepy.utils import log as _log

__all__ = [
    "armadillo",
    "chi",
    "cross_2d",
    "cube_void",
    "cross_3d_linear",
    "double_lattice",
    "ellips_v_oid",
    "hollow_cube",
    "hollow_octagon",
    "hollow_octagon_extrude",
    "inverse_cross_3d",
    "snappy",
    "tile_base",
    "Armadillo",
    "Chi",
    "Cross2D",
    "Cross3D",
    "Cross3DLinear",
    "CubeVoid",
    "DoubleLattice",
    "EllipsVoid",
    "HollowCube",
    "HollowOctagon",
    "HollowOctagonExtrude",
    "InverseCross3D",
    "Snappy",
    "TileBase",
    "by_dim",
    "everything",
    "show",
    "get",
]


def _summarize_tiles():
    """
    Get and save tile types
    """
    tile_types = {}
    d1 = {}
    d2 = {}
    d3 = {}

    # loop subclasses and classify
    for SubClass in TileBase.__subclasses__():
        key = SubClass.__qualname__
        # save types and sort with direction
        tile_types[key] = SubClass
        dim = SubClass.dim
        if dim == 1:
            d1[key] = SubClass
        elif dim == 2:
            d2[key] = SubClass
        elif dim == 3:
            d3[key] = SubClass

    return tile_types, d1, d2, d3


_tile_types, _1d, _2d, _3d = _summarize_tiles()


def by_dim(para_dim=None, dim=None):
    """
    Returns names of tiles that satisfies dimension inputs.
    Per default, it returns list of all the name of available tiles.

    Parameters
    ----------
    para_dim: int
    dim: int

    Returns
    -------
    tiles: dict<str, type>
        dict of tile names and types that fulfills dimension inputs.
    """
    # initialize pool of tiles
    pool = _tile_types.copy() if dim is None else eval(f"_{int(dim)}d").copy()

    # filter
    if para_dim is not None:
        para_dim = int(para_dim)
        filtered = {}
        for key, value in pool.items():
            if value.para_dim == para_dim:
                filtered[key] = value

        # overwrite pool with filtered
        pool = filtered

    if len(pool) == 0:
        _log.error(
            "Tiles does not exist with given dimension combination - ",
            f"(para_dim {para_dim}, dim {dim})",
        )

    return pool


def everything():
    """
    Returns all predefined tiles.

    Parameters
    ----------
    None

    Returns
    -------
    tiles: dict<str, type>
    """
    return _tile_types.copy()


def show(**kwargs):
    """
    Shows name and default tile (with default parameter values).

    Parameters
    ----------
    **kwargs: kwargs
      show options that splinepy.show() accepts

    Returns
    -------
    show: Any
      show() return based on kwargs. Look gus.show() documentations.
    """
    from splinepy import Multipatch, show

    to_show = []
    for key, value in _tile_types.items():
        to_show.append([key, Multipatch(value().create_tile()[0])])

    # turn off control points if kwargs doesn't have it
    if not any(k.startswith("control") for k in kwargs):
        kwargs["control_points"] = False

    return show(*to_show, **kwargs)


def get(key):
    """
    Returns initialized tile object based that corresponds to given str key.

    Parameters
    ----------
    key: str
      Name of a tile in str

    Returns
    -------
    tile: TileBase
      Initialized, derived tile.
    """
    if key in _tile_types:
        # return initialized as they don't expect any variables
        return _tile_types[key]()

    raise KeyError(f"{key}-tile does not exist.")



import numpy as np


def test_bspline_normalize_knot_vectors(bspline_2p2d):
    """ """
    # make "iga book" bspline
    bspline = bspline_2p2d

    ref = [kv.numpy() for kv in bspline.knot_vectors]

    for kv in bspline.knot_vectors:
        kv[:] = np.multiply(np.add(kv, 37), 1.78934).tolist()

    # normalize
    bspline.normalize_knot_vectors()

    for i, (ref_kv, kv) in enumerate(zip(ref, bspline.knot_vectors)):
        assert np.allclose(ref_kv, kv), f"{i}. para dim failed to normalize"


def test_nurbs_normalize_knot_vectors(nurbs_2p2d):
    """ """
    nurbs = nurbs_2p2d

    ref = [kv.numpy() for kv in nurbs.knot_vectors]

    # manipulate - this does not update cpp spline, but it doesn't matter
    for kv in nurbs.knot_vectors:
        kv[:] = np.multiply(np.add(kv, 37), 1.78934).tolist()

    # normalize
    nurbs.normalize_knot_vectors()

    for i, (ref_kv, kv) in enumerate(zip(ref, nurbs.knot_vectors)):
        assert np.allclose(ref_kv, kv), f"{i}. para dim failed to normalize"



"""splinepy/splinepy/settings.py
Convenient global variables.
"""


def __splinepy_name_to_type__():
    """Workaround to provide flexible string to type conversion without
    causing circular import
    """
    from splinepy import NURBS as _NURBS
    from splinepy import Bezier as _Bezier
    from splinepy import BSpline as _BSpline
    from splinepy import Multipatch as _Multipatch
    from splinepy import RationalBezier as _RationalBezier

    return {
        "Bezier": _Bezier,
        "RationalBezier": _RationalBezier,
        "BSpline": _BSpline,
        "NURBS": _NURBS,
        "Multipatch": _Multipatch,
    }


TOLERANCE = 1e-11
"""
Default tolerance for any operation that involves floats
"""

NTHREADS = 1
"""
Default number of threads.
"""

NAME_TO_TYPE = None
"""
String to Type conversion. Any function that may return different types should
use this. If you have derived classes, replace this with your own conversion
dict. In splinepy, this is set in __init__.
"""

CHECK_BOUNDS = True
"""
Bool to check bounds of queries if requested. Can be set to false to
accelerate process
"""



import numpy as _np

from splinepy._base import SplinepyBase as _SplinepyBase


class TileBase(_SplinepyBase):
    """
    Base class for tile objects
    """

    _dim = None
    _para_dim = None
    _evaluation_points = None
    _n_info_per_eval_point = None

    def __init__(self):
        if type(self) is TileBase:
            raise TypeError(
                "Can't initialize a TileBase class! Please use one of the "
                f"derived classes: {TileBase.__subclasses__()}"
            )

    @classmethod
    def _raise_if_not_set_else_return(cls, attr_name):
        attr = getattr(cls, attr_name, None)
        if attr is None and cls is not TileBase:
            raise NotImplementedError(
                f"Inherited Tile-types need to provide {attr_name}, see "
                "documentation."
            )
        return attr

    @classmethod
    @property
    def evaluation_points(cls):
        """Positions in the parametrization function to be evaluated when tile
        is constructed prior to composition.

        Parameters
        ----------
        None

        Returns
        -------
        evaluation_points : np.ndarray(6,3)
        """
        return cls._raise_if_not_set_else_return("_evaluation_points")

    @classmethod
    @property
    def dim(cls):
        """Returns dimensionality in physical space of the Microtile.

        Parameters
        ----------
        None

        Returns
        -------
        dim : int
        """
        return cls._raise_if_not_set_else_return("_dim")

    @classmethod
    @property
    def para_dim(cls):
        """Returns dimensionality in parametric space of the Microtile.

        Parameters
        ----------
        None

        Returns
        -------
        para_dim : int
        """
        return cls._raise_if_not_set_else_return("_para_dim")

    def check_params(self, params):
        """Checks if the parameters have the correct format and shape

        Parameters
        ----------
        params: np.ndarray
            the parameters for the tile

        Returns
        -------
        True: Boolean
        """
        # check if tuple

        if not (isinstance(params, _np.ndarray) and params.ndim == 2):
            raise TypeError("parameters must be two-dimensional np array")

        if not (
            (self._evaluation_points.shape[0] == params.shape[0])
            and (self._n_info_per_eval_point == params.shape[1])
        ):
            raise TypeError(
                f"Mismatch in parameter size, expected "
                f"{self._evaluation_points.shape[0]} x "
                f"{self._n_info_per_eval_point}"
            )

        return True

    def check_param_derivatives(self, derivatives):
        """Checks if all derivatives have the correct format and shape

        Parameters
        ----------
        derivatives: np.ndarray
            all derivatives as 3D array

        Returns
        -------
        True: Boolean
        """
        if derivatives is None:
            return False

        if not (
            isinstance(derivatives, _np.ndarray) and derivatives.ndim == 3
        ):
            raise TypeError("parameters must be three-dimensional np array")

        if not (
            (self._evaluation_points.shape[0] == derivatives.shape[0])
            and (self._n_info_per_eval_point == derivatives.shape[1])
        ):
            raise TypeError(
                f"Mismatch in parameter size, expected "
                f"{self._evaluation_points.shape[0]} x "
                f"{self._n_info_per_eval_point} x n_sensitivities"
            )

        return True

    def create_tile(self, **kwargs):
        """Tile creation interface for derived classes.
        All types should implement this.

        Parameters
        ----------
        **kwargs: kwargs

        Returns
        -------
        tiles: list
          List of splines that create a tile
        derivatives: list<list<Spline>> or None
          List of list of splines that represents parameter sensitivities.
          If it is not implemented, returns None.
        """
        raise NotImplementedError(
            f"create_tile() not implemented for {type(self)}"
        )



import gustaf as gus
import numpy as np
import vedo

import splinepy

# for consistent viewing for both dark and light modes
bg = "grey"

### 1.

# Initialize nurbs with any array-like input
nurbs = splinepy.NURBS(
    degrees=[2, 1],
    knot_vectors=[
        [0, 0, 0, 1, 1, 1],
        [0, 0, 1, 1],
    ],
    control_points=[
        [-1.0, 0.0],
        [-1.0, 1.0],
        [0.0, 1.0],
        [-2.0, 0.0],
        [-2.0, 2.0],
        [0.0, 2.0],
    ],
    weights=[
        [1.0],
        [2**-0.5],
        [1.0],
        [1.0],
        [2**-0.5],
        [1.0],
    ],
)

# vizusalize
nurbs.show(background=bg)


### 2.

# start with a copy of the original spline
modified = nurbs.copy()

# manipulate control points
# 1. all at once
modified.control_points /= 2.0
# 2. indexwise (flat indexing)
modified.control_points[[3, 4, 5]] *= [1.3, 2.0]
# 3. with grid-like indexing using multi_index helper
multi_index = modified.multi_index
modified.control_points[multi_index[0, 1]] = [-1.5, -0.3]
modified.control_points[multi_index[2, :]] += [2.0, 0.1]

# modified.show()  # visualize Nr. 1
view1 = modified.copy()

# elevate degrees and insert knots
modified.elevate_degrees([0, 1])
# modified.show()  # visualize Nr. 2
view2 = modified.copy()

modified.insert_knots(1, [0.5])
# modified.show()  # visualize Nr. 3
view3 = modified.copy()

splinepy.show(
    ["1", view1],
    ["2", view2],
    ["3", view3],
    background=bg,
)


### 3.

# first, create parametric coordinate queries
queries = [
    [0.1, 0.2],  # first query
    [0.4, 0.5],  # second query
    [0.1156, 0.9091],  # third query
]

# evaluate basis, spline and derivatives.
# for derivatives, specify order per parametric dimension.
basis = nurbs.basis(queries)
basis_derivative = nurbs.basis_derivative(queries, [1, 1])
physical_coordinates = nurbs.evaluate(queries)
physical_derivatives = nurbs.derivative(queries, [2, 0])

p_basis0 = nurbs.basis(queries, nthreads=2)
# or
splinepy.settings.NTHREADS = 3
p_basis1 = nurbs.basis(queries)

para_queries = []
phys_queries = []
colors = ["red", "yellow", "blue"]
for q, pc, c in zip(queries, physical_coordinates, colors):
    para_q, phys_q = gus.Vertices([q]), gus.Vertices([pc])
    para_q.show_options["c"] = c
    para_q.show_options["r"] = 10
    phys_q.show_options["c"] = c
    phys_q.show_options["r"] = 10
    para_queries.append(para_q)
    phys_queries.append(phys_q)

splinepy.show(
    ["Parametric", *para_queries, nurbs.create.parametric_view()],
    ["Physical", *phys_queries, nurbs],
    background=bg,
)

# see docs for options
para_coordinates = nurbs.proximities(physical_coordinates)
assert np.allclose(queries, para_coordinates)

(
    parametric_coordinates,
    physical_coordindates,
    physical_difference,
    distance,
    convergence_norm,
    first_derivatives,
    second_derivatives,
) = nurbs.proximities(physical_coordinates, return_verbose=True)


### 4.1

# basic splines
box = splinepy.helpme.create.box(1, 2, 3)  # length per dim
disk = splinepy.helpme.create.disk(outer_radius=3, inner_radius=2, angle=256)
torus = splinepy.helpme.create.torus(torus_radius=3, section_outer_radius=1.5)

splinepy.show(
    ["box", box],
    ["disk", disk],
    ["torus", torus],
    background=bg,
    resolutions=201,
)


# based on existing splines
extruded = nurbs.create.extruded(extrusion_vector=[1, 2, 3])
revolved = nurbs.create.revolved(axis=[1, 0, 0], center=[-1, -1, 0], angle=50)

splinepy.show(["extruded", extruded], ["revolved", revolved], background=bg)

### 4.2

# extract meshes as gustaf objects
control_mesh = nurbs.extract.control_mesh()
control_points = nurbs.extract.control_points()
mesh = nurbs.extract.faces([201, 33])  # specify sample resolutions

control_mesh_as_edges = control_mesh.to_edges()
control_mesh_as_edges.show_options.update(c="black", lw=8)
splinepy.show(
    ["control mesh", control_mesh_as_edges],
    ["control points", control_points],
    ["spline", mesh],
    background=bg,
)

# extract splines
boundaries = nurbs.extract.boundaries()
partial = nurbs.extract.spline(0, [0.5, 0.78])
partial_partial = nurbs.extract.spline(0, [0.1, 0.3]).extract.spline(
    1, [0.65, 0.9]
)
bases = nurbs.extract.bases()  # basis functions as splines
# insert knots to increase number of bezier patches
inserted = nurbs.copy()
inserted.insert_knots(0, [0.13, 0.87])
beziers_patches = inserted.extract.beziers()

splinepy.show(
    ["boundaries and part of splines", boundaries, partial, partial_partial],
    ["beziers", beziers_patches],
    ["bases", bases],
    background=bg,
)

### 4.3
# create gustaf mesh using extract.spline()
# or use gustaf's io functions (gustaf.io)
mesh = splinepy.helpme.create.torus(2, 1).extract.faces([100, 100, 100])

# initialize ffd and move control points
ffd = splinepy.FFD(mesh=mesh)
multi_index = ffd.spline.multi_index
ffd.spline.control_points[multi_index[-1, :, -1]] += [3, 0.5, 0.1]

cam = {
    "position": (-9.96434, -7.98279, 7.95655),
    "focal_point": (0.361918, -0.129347, 0.348385),
    "viewup": (0.440934, 0.254149, 0.860805),
    "roll": 72.6622,
    "distance": 15.0397,
    "clipping_range": (7.14564, 27.5698),
}

ffd.show(background=bg, cam=cam)

# get deformed mesh - FFD.mesh attribute deforms mesh before returning
deformed = ffd.mesh


### 4.4
data = [
    [-0.955, 0.293],
    [-0.707, 0.707],
    [-0.293, 0.955],
    [-1.911, 0.587],
    [-1.414, 1.414],
    [-0.587, 1.911],
]

curve, residual_curve = splinepy.helpme.fit.curve(data, degree=2)
# you can also use any existing spline's basis
surface, residual_surface = splinepy.helpme.fit.surface(
    data, size=[3, 2], fitting_spline=nurbs
)

# set visuals for data
d = gus.Vertices(data)
d.show_options.update(c="blue", r=15)

splinepy.show(
    ["curve fit", d, curve],
    ["surface fit", d, surface],
    background=bg,
)


### 4.5
# create solution spline
solution_field = nurbs.create.embedded(1)

# refine
solution_field.elevate_degrees([0, 1])
solution_field.uniform_refine(n_knots=[4, 4])

# create matrix using mapper
# collocation points at greville abcissae
mapper = solution_field.mapper(reference=nurbs)
laplacian, support = mapper.basis_laplacian_and_support(
    solution_field.greville_abscissae()
)
laplacian_matrix = splinepy.utils.data.make_matrix(
    laplacian,
    support,
    n_cols=solution_field.control_points.shape[0],
    as_array=True,
)

# for this one, we want zero to the mittle value to be zero to highlight
# sparse matrix
vmin = laplacian_matrix.min()
vmax = laplacian_matrix.max()
v_abs = max(vmin, vmax)
vmin = v_abs * np.sign(vmin)
vmax = v_abs * np.sign(vmax)

vedo_mat = vedo.pyplot.matrix(
    laplacian_matrix, cmap="seismic", vmin=vmin, vmax=vmax
)
splinepy.show(vedo_mat, background=bg)


### 5.

splinepy.microstructure.tiles.show(alpha=0.7, background=bg)

# get specific dimensions as dict
para_2_dim_2 = splinepy.microstructure.tiles.by_dim(para_dim=2, dim=2)
dim_2 = splinepy.microstructure.tiles.by_dim(dim=2)


# create microstructure generator
microstructure = splinepy.Microstructure()
# set outer spline and a (micro) tile
microstructure.deformation_function = nurbs
microstructure.microtile = splinepy.microstructure.tiles.get("Cross2D")
# tiling determines tile resolutions within each bezier patch
microstructure.tiling = [5, 3]

microstructure.show(background=bg)

# extract only generated parts as multipatch
generated = microstructure.create()


### 6.
# use previously generated microtiles
interface_info_array = generated.interfaces


# Mark boundaries to set boundary conditions
# In case of micro structure, you can use outer spline's boundary
def is_left_bdr(x):
    left = nurbs.extract.boundaries()[3]
    return (left.proximities(x, return_verbose=True)[3] < 1e-8).ravel()


generated.boundary_from_function(is_left_bdr, boundary_id=5)

splinepy.show(
    ["All", generated],
    ["Boundaries", generated.boundary_multipatch()],
    ["Boundary 5", generated.boundary_multipatch(5)],
    control_points=False,
    background=bg,
)

# export for the solver
splinepy.io.gismo.export("microstructure.xml", generated)


### 7.
# export
splinepy.io.mfem.export("quarter_circle.mesh", nurbs)

# load
quarter_circle = splinepy.io.mfem.load("quarter_circle.mesh")

# svg export
splinepy.io.svg.export("nurbs.svg", nurbs, background=bg)



import numpy as np

import splinepy


def test_greville_points_bezier(bezier_3p3d):
    """
    test greville point calculation for bezier
    """
    greville_points = splinepy.utils.data.cartesian_product(
        [
            np.linspace(0, 1, bezier_3p3d.degrees[i] + 1)
            for i in range(bezier_3p3d.para_dim)
        ]
    )

    assert np.allclose(greville_points, bezier_3p3d.greville_abscissae())


def test_greville_points_rational_bezier(rational_bezier_3p3d):
    """
    test greville point calculation for rational bezier
    """
    greville_points = splinepy.utils.data.cartesian_product(
        [
            np.linspace(0, 1, rational_bezier_3p3d.degrees[i] + 1)
            for i in range(rational_bezier_3p3d.para_dim)
        ]
    )
    assert np.allclose(
        greville_points, rational_bezier_3p3d.greville_abscissae()
    )


def test_greville_points_bspline(bspline_3p3d):
    """
    test greville point calculation for bspline
    """
    # Modify some splines
    bspline_3p3d.insert_knots(0, [0.2, 0.7])
    bspline_3p3d.insert_knots(1, [0.2, 0.5, 0.7])
    bspline_3p3d.insert_knots(2, [0.5])

    # Test non-unifrom Splines
    greville_points = splinepy.utils.data.cartesian_product(
        [
            np.convolve(
                bspline_3p3d.knot_vectors[i][1:-1],
                np.ones(bspline_3p3d.degrees[i]),
            )
            for i in range(bspline_3p3d.para_dim)
        ]
    )
    assert np.allclose(greville_points, bspline_3p3d.greville_abscissae())


def test_greville_points_nurbs(nurbs_3p3d):
    """
    test greville point calculation for nurbs
    """
    nurbs_3p3d.elevate_degrees([0, 0, 1, 2])
    nurbs_3p3d.insert_knots(0, [0.25])
    # Test Uniform Types
    greville_points = splinepy.utils.data.cartesian_product(
        [
            np.convolve(
                nurbs_3p3d.knot_vectors[i][1:-1],
                np.ones(nurbs_3p3d.degrees[i]) / nurbs_3p3d.degrees[i],
                mode="valid",
            )
            for i in range(nurbs_3p3d.para_dim)
        ]
    )
    assert np.allclose(greville_points, nurbs_3p3d.greville_abscissae())


def test_greville_with_duplicate_points():
    """Tests if duplicate points are filtered out, as is required for the
    construction of c^(-1) spline"""
    # Construct c^(-1) spline
    a = splinepy.BSpline(
        degrees=[2],
        knot_vectors=[[0, 0, 0, 0.25, 0.5, 0.5, 0.5, 0.75, 1, 1, 1]],
        control_points=np.ones((8, 2)),
    )

    assert np.allclose(
        a.greville_abscissae(duplicate_tolerance=-1.0).ravel(),
        [0, 0.125, 0.375, 0.5, 0.5, 0.625, 0.875, 1.0],
    )

    assert np.allclose(
        a.greville_abscissae(
            duplicate_tolerance=splinepy.settings.TOLERANCE
        ).ravel(),
        [0, 0.125, 0.375, 0.4375, 0.5625, 0.625, 0.875, 1.0],
    )



import gustaf as gus
import numpy as np

import splinepy

if __name__ == "__main__":
    line = splinepy.helpme.create.line(
        np.array([[0, 0, 0], [2, 5, 0], [4, 4, 2]])
    )
    rect = splinepy.helpme.create.box(5, 3)
    box = splinepy.helpme.create.box(3, 2, 4)
    pyramid = splinepy.helpme.create.pyramid(1, 1, 2)

    gus.show(
        ["Line", line],
        ["Rectangle", rect],
        ["Box", box],
        ["Pyramid", pyramid],
        title="Rectangular objects",
        resolution=50,
    )

    circ = splinepy.helpme.create.circle(5)
    disk1 = splinepy.helpme.create.disk(3, angle=100, n_knot_spans=4)
    disk2 = splinepy.helpme.create.disk(
        5, inner_radius=1.0, angle=360, n_knot_spans=10
    )

    gus.show(["Line Circle", circ], ["Disk section", disk1], ["Disk", disk2])

    cone = splinepy.helpme.create.cone(5, 10, angle=180)
    gus.show(["Cone", cone])

    torus = splinepy.helpme.create.torus(10, 2)
    torus2 = splinepy.helpme.create.torus(
        5, 2, section_inner_radius=0.5, torus_angle=100, section_angle=210
    )

    gus.show(["Torus", torus], ["Torus section", torus2], resolution=50)

    sphere = splinepy.helpme.create.sphere(3)
    gus.show(["Sphere", sphere], resolution=50)



import gustaf as gus
import numpy as np

import splinepy


def test_uniform_query(np_rng):
    """
    Test uniform query against gustaf's vertice creation
    """
    for i in range(1, 11):
        l_bound = np_rng.random(i)
        bounds = [l_bound, l_bound + 1]
        res = np_rng.integers([3] * i) + 2  # minimum 2
        gus_q = gus.create.vertices.raster(bounds, res).vertices
        spp_q = splinepy.utils.data.uniform_query(bounds, res)

        assert np.allclose(gus_q, spp_q)



import os
import tempfile
from sys import version as python_version

import numpy as np
import pytest

import splinepy


@pytest.fixture
def gismo_2D_multipatch():
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bez_el0.elevate_degrees(0)
    bsp_el2.elevate_degrees(0)

    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Init multipatch
    multipatch = splinepy.Multipatch(
        splines=[bez_el0, rbz_el1, bsp_el2, nur_el3]
    )

    # Define some functions for boundary identification
    def is_bottom(x):
        return x[:, 0] < 0.01

    def is_top(x):
        return x[:, 0] > 1.99

    # Add boundary
    multipatch.boundary_from_function(is_bottom)
    multipatch.boundary_from_function(is_top)

    return multipatch


@pytest.fixture
def gismo_multipatch_3D():
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1, 1],
        control_points=[
            [1, 0, 0],
            [2, 0, 0],
            [1, 1, 0],
            [2, 1, 0],
            [1, 0, 1],
            [2, 0, 1],
            [1, 1, 1],
            [2, 1, 1],
        ],
        weights=[1] * 8,
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1, 1],
        control_points=[
            [0, 1, 0],
            [1, 1, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 1, 1],
            [1, 1, 1],
            [0, 2, 1],
            [1, 2, 1],
        ],
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1, 1],
        control_points=[
            [1, 1, 0],
            [2, 1, 0],
            [1, 2, 0],
            [2, 2, 0],
            [1, 1, 1],
            [2, 1, 1],
            [1, 2, 1],
            [2, 2, 1],
        ],
        weights=[1] * 8,
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )

    # Make it more tricky
    bez_el0.elevate_degrees(0)
    bsp_el2.elevate_degrees(0)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])
    for s in [bez_el0, bsp_el2, nur_el3, rbz_el1]:
        s.elevate_degrees(2)

    # Define some functions for boundary identification
    def is_bottom(x):
        return x[:, 0] < 0.01

    def is_top(x):
        return x[:, 0] > 1.99

    # Add boundary
    multipatch = splinepy.Multipatch(
        splines=[bez_el0, bsp_el2, nur_el3, rbz_el1]
    )
    multipatch.boundary_from_function(is_bottom)
    multipatch.boundary_from_function(is_top)
    return multipatch


def test_gismo_export_2D(
    to_tmpf, gismo_2D_multipatch, are_stripped_lines_same
):
    """
    Test gismo export routine for 2D
    """
    ###########
    # 2D Mesh #
    ###########
    # Define some splines

    # Test Output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=gismo_2D_multipatch,
            indent=False,
            labeled_boundaries=False,
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__)
            + "/../data/gismo_noindent_nolabels_ascii_2d.xml"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )


def test_gismo_export_2D_indented(
    gismo_2D_multipatch, to_tmpf, are_stripped_lines_same
):
    # for python version > 3.9, test indented version
    if int(python_version.split(".")[1]) >= 9:
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = to_tmpf(tmpd)
            splinepy.io.gismo.export(
                tmpf,
                multipatch=gismo_2D_multipatch,
                indent=True,
                labeled_boundaries=False,
            )

            with open(tmpf) as tmp_read, open(
                os.path.dirname(__file__)
                + "/../data/gismo_indent_nolabels_ascii_2d.xml"
            ) as base_file:
                assert are_stripped_lines_same(
                    base_file.readlines(), tmp_read.readlines(), True
                )


def test_gismo_export_2D_labels(
    gismo_2D_multipatch, to_tmpf, are_stripped_lines_same
):
    ########################
    # 2D Mesh - new format #
    ########################
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=gismo_2D_multipatch,
            indent=False,
            labeled_boundaries=True,
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__)
            + "/../data/gismo_noindent_labels_ascii_2d.xml"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )


def test_gismo_export_2D_labels_indented(
    gismo_2D_multipatch, to_tmpf, are_stripped_lines_same
):
    # for python version > 3.9, test indented version
    if int(python_version.split(".")[1]) >= 9:
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = to_tmpf(tmpd)
            splinepy.io.gismo.export(
                tmpf,
                multipatch=gismo_2D_multipatch,
                indent=True,
                labeled_boundaries=True,
            )

            with open(tmpf) as tmp_read, open(
                os.path.dirname(__file__)
                + "/../data/gismo_indent_labels_ascii_2d.xml"
            ) as base_file:
                assert are_stripped_lines_same(
                    base_file.readlines(), tmp_read.readlines(), True
                )


def test_gismo_export_3D(
    gismo_multipatch_3D, to_tmpf, are_stripped_lines_same
):
    # Test Also 3D Meshes

    # Test output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=gismo_multipatch_3D,
            indent=False,
            labeled_boundaries=False,
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__)
            + "/../data/gismo_noindent_nolabels_ascii_3d.xml"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )


def test_gismo_export_3D_indented(
    gismo_multipatch_3D, to_tmpf, are_stripped_lines_same
):
    # for python version > 3.9, test indented version
    if int(python_version.split(".")[1]) >= 9:
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = to_tmpf(tmpd)
            splinepy.io.gismo.export(
                tmpf,
                multipatch=gismo_multipatch_3D,
                indent=True,
                labeled_boundaries=False,
            )

            with open(tmpf) as tmp_read, open(
                os.path.dirname(__file__)
                + "/../data/gismo_indent_nolabels_ascii_3d.xml"
            ) as base_file:
                assert are_stripped_lines_same(
                    base_file.readlines(), tmp_read.readlines(), True
                )


def test_gismo_import(to_tmpf, are_splines_equal):
    """
    Test gismo export routine
    """
    # Define some splines
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 0], [1, 0], [0, 1], [1, 1]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test Output against input
    multipatch_geometry = splinepy.Multipatch([bsp_el2, nur_el3])
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=multipatch_geometry,
            indent=False,
            labeled_boundaries=False,
        )
        multipatch_geometry_loaded = splinepy.io.gismo.load(
            tmpf, load_options=False
        )
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(
                multipatch_geometry.patches,
                multipatch_geometry_loaded.patches,
            )
        )

        assert np.allclose(
            multipatch_geometry.interfaces,
            multipatch_geometry_loaded.interfaces,
        )

        # Now with modified boundaries
        multipatch_geometry.boundaries_from_continuity()
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=multipatch_geometry,
            indent=False,
            labeled_boundaries=True,
        )
        multipatch_geometry_loaded = splinepy.io.gismo.load(
            tmpf, load_options=False
        )
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(
                multipatch_geometry.patches,
                multipatch_geometry_loaded.patches,
            )
        )

        assert np.allclose(
            multipatch_geometry.interfaces,
            multipatch_geometry_loaded.interfaces,
        )


def test_gismo_import_with_options(to_tmpf, are_splines_equal):
    """
    Test gismo export routine
    """
    if int(python_version.split(".")[1]) < 8:
        splinepy.utils.log.info(
            "gismo export is only tested here from python3.8+. "
            "Skipping test, because current version is: "
            f"{python_version}"
        )
        return True

    # Define some splines
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 0], [1, 0], [0, 1], [1, 1]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test Output against input
    multipatch_geometry = splinepy.Multipatch([bsp_el2, nur_el3])

    # Set some options
    gismo_options = [
        {
            "tag": "OptionNumber1",
            "attributes": {"Mambo": "No. 5", "id": "5"},
            "text": "One, two, three, four, five\nEverybody in the car, "
            "so come on, let's ride",
            "children": [
                {
                    "tag": "AnotherOne",
                    "text": "0 ,0",
                    "attributes": {},
                    "children": [],
                }
            ],
        }
    ]
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=multipatch_geometry,
            indent=False,
            labeled_boundaries=False,
            options=gismo_options,
        )
        (
            multipatch_geometry_loaded,
            gismo_options_loaded,
        ) = splinepy.io.gismo.load(tmpf, load_options=True)
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(
                multipatch_geometry.patches,
                multipatch_geometry_loaded.patches,
            )
        )

        assert np.allclose(
            multipatch_geometry.interfaces,
            multipatch_geometry_loaded.interfaces,
        )

        assert gismo_options_loaded == gismo_options


def test_gismo_io_binary(to_tmpf, are_stripped_lines_same, are_splines_equal):
    """Test the base64 io-routines"""
    # We test this with just one (big, 3D) spline
    nurbs = splinepy.NURBS(
        degrees=[1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1], [0, 0, 1, 1]],
        control_points=np.ones((8, 3)),
        weights=np.ones((8, 1)),
    )
    nurbs.elevate_degrees([0, 1, 2, 2])
    np_rng = np.random.RandomState(19284918)
    for i in range(3):
        nurbs.insert_knots(i, np_rng.random(4))

    # Randomize points
    nurbs.cps = np_rng.random(nurbs.cps.shape)
    nurbs.weights = np_rng.random(nurbs.weights.shape)

    # Create a multipatch geometry
    multipatch_geometry = splinepy.Multipatch([nurbs])

    # Export
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.gismo.export(
            tmpf,
            multipatch=multipatch_geometry,
            indent=False,
            labeled_boundaries=False,
            as_base64=True,
        )
        (
            multipatch_geometry_loaded,
            gismo_options_loaded,
        ) = splinepy.io.gismo.load(tmpf, load_options=True)

        with open(tmpf) as tmp_read, open(
            os.path.dirname(os.path.dirname(__file__))
            + "/data/gismo_noindent_nolabels_b64_3d.xml"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

        assert all(
            are_splines_equal(a, b)
            for a, b in zip(
                multipatch_geometry.patches,
                multipatch_geometry_loaded.patches,
            )
        )

        assert np.allclose(
            multipatch_geometry.interfaces,
            multipatch_geometry_loaded.interfaces,
        )



import pathlib

from splinepy.io import gismo as _gismo
from splinepy.io import iges as _iges
from splinepy.io import irit as _irit
from splinepy.io import json as _json
from splinepy.io import mfem as _mfem
from splinepy.io import npz as _npz
from splinepy.io import svg as _svg

extension_to_io = {
    ".xml": _gismo,
    ".iges": _iges,
    ".igs": _iges,
    ".itd": _irit,
    ".irit": _irit,
    ".json": _json,
    ".mesh": _mfem,
    ".npz": _npz,
    ".svg": _svg,
}


def export(fname, splinepy_obj):
    """
    Exports a splinepy object to a file.

    Tries to find out the correct exporter from the file extension. If the
    extension is not supported, a ValueError is raised.

    Parameters
    ----------
    fname : str
      Export Filename
    splinepy_obj: Splinepy object
      Splinepy object to be exported.

    Returns
    -------
    output_dict : dict
      Dictornay data written into file

    """
    # Determine file extension
    extension = pathlib.Path(fname).suffix
    if extension not in extension_to_io:
        raise ValueError("Extension not supported")

    # Export
    extension_to_io[extension].export(fname, splinepy_obj)


def load(fname):
    """
    Loads a splinepy object from a file.

    Tries to find out the correct importer from the file extension. If the
    extension is not supported, a ValueError is raised.

    Parameters
    ----------
    fname : str
      Filename to load

    Returns
    -------
    splinepy_obj : Splinepy object
      Splinepy object loaded from file

    """
    # Determine file extension
    extension = pathlib.Path(fname).suffix
    if extension not in extension_to_io:
        raise ValueError("Extension not supported")

    # Load
    return extension_to_io[extension].load(fname)



import numpy as np
import pytest

# fixtures for test_elevate_degree
all_2p2d_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)

# fixtures for test_reduce_degree
nurbs_and_bspline = (
    "nurbs_2p2d",
    "bspline_2p2d",
)


@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_elevate_degree(splinetype, request, np_rng, are_items_close):
    """Test the order elevation function (.elevate_degrees())."""

    # reference solution
    bspline_ref_kv = [
        [0, 0, 0, 0, 0.5, 0.5, 1, 1, 1, 1],
        [0, 0, 0, 1, 1, 1],
    ]
    nurbs_ref_kv = [
        [0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 1, 1],
    ]

    # define the spline and a copy as reference
    spline = request.getfixturevalue(splinetype)
    ref_spline = spline.copy()

    # elevate order
    spline.elevate_degrees(0)

    # test degrees and knot-vectors
    if splinetype == "bspline_2p2d":
        assert np.allclose(spline.degrees, [3, 2])
        assert are_items_close(spline.knot_vectors, bspline_ref_kv)
    elif splinetype == "nurbs_2p2d":
        assert np.allclose(spline.degrees, [3, 1])
        assert are_items_close(spline.knot_vectors, nurbs_ref_kv)
    else:
        assert np.allclose(spline.degrees, [3, 1])

    # use random query points
    q2D = np_rng.random((10, 2))

    # test evaluation
    assert np.allclose(spline.evaluate(q2D), ref_spline.evaluate(q2D))


@pytest.mark.parametrize("splinetype", nurbs_and_bspline)
def test_reduce_degree(splinetype, request, are_items_close, np_rng):
    """Test the function .reduce_degrees.
    This test also depends on the function .elevate_degrees!"""

    # define the spline and a copy as reference
    spline = request.getfixturevalue(splinetype)
    ref_spline = spline.copy()

    # elevate and reduce order
    spline.elevate_degrees(0)
    spline.reduce_degrees(0)

    # test degrees
    assert np.allclose(spline.degrees, ref_spline.degrees)

    # test knot_vectors
    assert are_items_close(spline.knot_vectors, ref_spline.knot_vectors)

    # use random query points
    q2D = np_rng.random((10, 2))

    # test evaluation
    assert np.allclose(spline.evaluate(q2D), ref_spline.evaluate(q2D))



import numpy as np
import pytest

# used fixtures for parametrization
all_2D_splinetypes = (
    "bspline_2p2d",
    "nurbs_2p2d",
    "bezier_2p2d",
    "rational_bezier_2p2d",
)


@pytest.mark.parametrize("splinetype", all_2D_splinetypes)
def test_c_contiguous_array_input(splinetype, request):
    """Test whether cpp sides receives contiguous array"""

    # define spline and its properties
    spl = request.getfixturevalue(splinetype)
    properties = spl.todict()

    # make f contiguous
    f_contig_orig = np.asarray(properties["control_points"], order="F")
    assert not f_contig_orig.flags["C_CONTIGUOUS"]
    assert f_contig_orig.flags["F_CONTIGUOUS"]

    # set non contiguous array as prop
    properties["control_points"] = f_contig_orig

    # check round_trip cps
    round_trip_cps = spl.cps

    # they should be the same
    assert np.allclose(round_trip_cps, properties["control_points"])

    # test setter
    # make sure this array is still not c contiguous
    assert not f_contig_orig.flags["C_CONTIGUOUS"]

    spl.cps = f_contig_orig  # setter

    assert spl.cps.flags["C_CONTIGUOUS"]
    assert np.allclose(spl.cps, properties["control_points"])



"""
Test composition sensitivity, i.e., the derivative concerning the
deformation function's control points.
"""

import numpy as np

import splinepy


def test_composition():
    # Here we only try 2D compositions (surface-line is tested in bezman)
    # Init Splines to be tested
    surface_polynomial = splinepy.Bezier(
        degrees=[1, 2],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2], [1, 3], [2, 3]],
    )
    surface_rational = splinepy.RationalBezier(
        degrees=[2, 1],
        control_points=[[1, 1], [2, 1], [3, 2], [1, 2], [2, 2], [3, 3]],
        weights=[1.0, 0.8, 1.0, 1.0, 0.8, 1.0],
    )
    inner_polynomial = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[[0.2, 0.0], [1.0, 0.2], [0.0, 0.8], [0.8, 1.0]],
    )
    inner_rational = splinepy.RationalBezier(
        degrees=[2, 1],
        control_points=[
            [0.2, 0.0],
            [0.5, 0.5],
            [1.0, 0.2],
            [0.0, 0.8],
            [0.5, 0.5],
            [0.8, 1.0],
        ],
        weights=[1, 0.5, 1, 1, 0.5, 1],
    )

    dx = 1e-4
    tolerance = 1e-4

    def test_splines(outer, inner):
        # Determine base line
        composed, composed_sensitivities = outer.compose(
            inner, compute_sensitivities=True
        )

        # Check against FD approximation
        for i in range(outer.cps.shape[0]):
            for j in range(outer.cps.shape[1]):
                outer.cps[i, j] += dx

                # FD values
                composed_dx = outer.compose(inner)

                # FD
                derivative_fd = (composed_dx.cps - composed.cps) / dx
                expected = np.zeros(composed.cps.shape)
                expected[:, j] = composed_sensitivities[i].cps.flatten()

                assert np.allclose(derivative_fd, expected, atol=tolerance)

                outer.cps[i, j] -= dx

    test_splines(surface_polynomial, inner_polynomial)
    test_splines(surface_polynomial, inner_rational)
    test_splines(surface_rational, inner_polynomial)
    test_splines(surface_rational, inner_rational)


def test_composition_sensitivities_on_bsplines(bspline_2p2d):
    """Combine Composition sensitivities with BSpline extraction"""

    # Initialize outer function
    bspline = bspline_2p2d

    # Initialize inner function
    inner_function = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[
            [0.5, 0.0],
            [1.0, 0.5],
            [0.0, 0.5],
            [0.5, 1.0],
        ],
    )

    # Initialize values for finite differences
    bspline_dx = bspline.copy()
    bspline.insert_knots(1, 0.5)
    bspline_dx.insert_knots(1, 0.5)
    dx = 1e-5

    # Perform computations
    extract_beziers = bspline.extract_bezier_patches()
    extraction_matrices = bspline.knot_insertion_matrix(beziers=True)

    composed_der_ctps = []
    beziers = []
    for bez, mat in zip(extract_beziers, extraction_matrices):
        # Composition
        composed, derivatives = bez.compose(
            inner_function, compute_sensitivities=True
        )
        # Form matrix
        deriv_points = np.hstack([der.cps for der in derivatives])

        # Compute derivatives
        composed_der_ctps.append(deriv_points @ mat)
        beziers.append(composed)

    for cps in range(bspline.cps.shape[0]):
        for dim in range(bspline.cps.shape[1]):
            # Modify dx spline
            bspline_dx.cps[cps, dim] += dx

            # Extract Beziers
            extract_beziers_dx = bspline_dx.extract_bezier_patches()
            for bez, bez_dx, comps in zip(
                beziers, extract_beziers_dx, composed_der_ctps
            ):
                # Compose finite differences spline
                composed_dx = bez_dx.compose(inner_function)
                fd_ctps = (composed_dx.cps - bez.cps) / dx

                assert np.allclose(fd_ctps[:, dim], comps[:, cps])

            # Reset dx spline
            bspline_dx.cps[cps, dim] -= dx


def test_sum(np_rng, bezier_2p2d):
    # Create two splines
    bezier1 = bezier_2p2d
    bezier2 = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[
            [0.5, 0.0],
            [1.0, 0.5],
            [0.0, 0.5],
            [0.5, 1.0],
        ],
    )
    bezier_sum = bezier1 + bezier2

    # Create queries
    n_test_points = 10
    para_dim = 2
    queries = np_rng.random((n_test_points, para_dim))

    assert np.allclose(
        bezier2.evaluate(queries) + bezier1.evaluate(queries),
        bezier_sum.evaluate(queries),
    )


def test_multiply(np_rng, bezier_2p2d):
    # Create two splines
    bezier1 = bezier_2p2d
    bezier2 = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[
            [0.0],
            [0.5],
            [0.5],
            [1.0],
        ],
    )
    bezier_prod = bezier1 * bezier2

    # Create queries
    n_test_points = 10
    para_dim = 2
    queries = np_rng.random((n_test_points, para_dim))

    assert np.allclose(
        bezier2.evaluate(queries) * bezier1.evaluate(queries),
        bezier_prod.evaluate(queries),
    )


def test_close_form_derivative(np_rng):
    # Construct a random spline
    degrees = [5, 5, 5]
    orders = [3, 2, 4]
    bezier = splinepy.Bezier(
        degrees=degrees,
        control_points=np_rng.random((np.prod(np.array(degrees) + 1), 3)),
    )
    close_form = bezier.derivative_spline(orders)

    # Create queries
    n_test_points = 100
    para_dim = bezier.para_dim
    queries = np_rng.random((n_test_points, para_dim))

    assert np.allclose(
        bezier.derivative(queries, orders),
        close_form.evaluate(queries),
    )

    # Repeat for rationals (low orders required, because orders square with
    # every derivative)
    degrees = [2, 1, 2]
    orders = [1, 0, 1]
    n_cps = np.prod(np.array(degrees) + 1)
    rationals = splinepy.RationalBezier(
        degrees=degrees,
        control_points=np_rng.random((n_cps, 3)),
        weights=np.abs(np_rng.random(n_cps)),
    )
    close_form = rationals.derivative_spline(orders)

    # Create queries
    n_test_points = 100
    para_dim = rationals.para_dim
    queries = np_rng.random((n_test_points, para_dim))

    assert np.allclose(
        rationals.derivative(queries, orders),
        close_form.evaluate(queries),
    )



"""Test Orientation between adjacent spline patches
         2--3
         |C |
         0--1
    0--1 2--3 1--3
    |D | |A | | B|
    2--3 0--1 0--2
         1--0
    |    |E |
    O-   3--2
"""

import numpy as np

import splinepy


def test_orientation():
    # Init Splines to be tested
    a_s = splinepy.Bezier(
        degrees=[1, 1], control_points=[[1, 1], [2, 1], [1, 2], [2, 2]]
    )
    b_s = splinepy.Bezier(
        degrees=[1, 1], control_points=[[2, 1], [2, 2], [3, 1], [3, 2]]
    )
    c_s = splinepy.Bezier(
        degrees=[1, 1], control_points=[[1, 2], [2, 2], [1, 3], [2, 3]]
    )
    d_s = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 2], [1, 2], [0, 1], [1, 1]]
    )
    e_s = splinepy.Bezier(
        degrees=[1, 1], control_points=[[2, 1], [1, 1], [2, 0], [1, 0]]
    )
    # Expected Values
    expected_interfaces = [
        [13, 6, 18, 10],
        [-1, -1, 1, -1],
        [-1, -1, 3, -1],
        [-1, 0, -1, -1],
        [-1, -1, 2, -1],
    ]

    expected_orientations = [
        [0, 0, 3, 1, 0, 1, 1, 0],
        [0, 1, 1, 2, 1, 0, 1, 1],
        [0, 2, 4, 2, 0, 1, 0, 0],
        [0, 3, 2, 2, 0, 1, 1, 1],
    ]

    # Provide connectivity data
    mp = splinepy.Multipatch([a_s, b_s, c_s, d_s, e_s])

    # First compare interfaces to what is expected
    interfaces = mp.interfaces
    assert np.all(interfaces == expected_interfaces)

    # Then check orientations
    orientations = mp.interface_orientations()
    assert np.all(orientations == expected_orientations)



r"""
In this file we will prototype a galerkin IGA example (in the simplest way),
that shows how to implement a Poisson problem in the form
.. math::
        -\Delta u = f
with source term f=1

The refinement will only be applied to the solution field, to make the
calculations more efficient
"""

import numpy as np

import splinepy as sp

# Test Case
n_refine = 15


# Source Function
def source_function(x):
    return np.ones(x.shape[0])


# Auxiliary function to map positions into the element
def map_positions(positions, x_min, x_max, y_min, y_max):
    mapped_positions = np.empty(positions.shape, dtype=np.float64)
    mapped_positions[:, 0] = (x_max - x_min) * 0.5 * positions[:, 0] + (
        x_max + x_min
    ) * 0.5
    mapped_positions[:, 1] = (y_max - y_min) * 0.5 * positions[:, 0] + (
        y_max + y_min
    ) * 0.5
    return mapped_positions


# Define the Geometry
geometry = sp.BSpline(
    degrees=[2, 2],
    control_points=[
        [0.0, 0.0],
        [1.0, 0.5],
        [2.0, 0.2],
        [0.5, 1.5],
        [1.0, 1.5],
        [1.5, 1.5],
        [0.0, 3.0],
        [1.0, 2.5],
        [2.0, 3.0],
    ],
    knot_vectors=[[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 1]],
)

# Define the solution field
solution_field = sp.BSpline(
    degrees=geometry.degrees,
    control_points=np.ones((geometry.control_points.shape[0], 1)),
    knot_vectors=geometry.knot_vectors,
)


# Use refinement
solution_field.elevate_degrees([0, 1, 0, 1])
new_knots = np.linspace(1 / n_refine, 1, n_refine, endpoint=False)
solution_field.insert_knots(0, new_knots)
solution_field.insert_knots(1, new_knots)

# Retrieve integration points and weights
max_order = int(np.max(solution_field.degrees))
positions, weights = np.polynomial.legendre.leggauss(deg=max_order)
positions = sp.utils.data.cartesian_product(
    [positions for _ in range(geometry.para_dim)]
)
weights = sp.utils.data.cartesian_product(
    [weights for _ in range(geometry.para_dim)]
)
weights = np.prod(weights, axis=1)

# Assemble individual elements
ukv = solution_field.unique_knots
n_dofs = solution_field.control_points.shape[0]
system_matrix = np.zeros((n_dofs, n_dofs))
system_rhs = np.zeros(n_dofs)
mapper = solution_field.mapper(reference=geometry)

# Element Loop
for i in range(len(ukv[0]) - 1):
    for j in range(len(ukv[1]) - 1):
        # Auxiliary values
        mapped_positions = map_positions(
            positions,
            ukv[0][i],
            ukv[0][i + 1],
            ukv[1][j],
            ukv[1][j + 1],
        )
        det_jacs = np.linalg.det(geometry.jacobian(mapped_positions))

        # RHS
        # q : quadrature point | d: dim
        bf, support = solution_field.basis_and_support(mapped_positions)
        assert np.all(support[0, :] == support)
        system_rhs[support[0, :]] += np.einsum(
            "qj,q,q->j", bf, weights, det_jacs, optimize=True
        )

        # LHS
        bf_jacobian, support = mapper.basis_gradient_and_support(
            mapped_positions
        )
        assert np.all(support[0, :] == support)
        local_matrix = np.einsum(
            "qid,qjd,q,q->ij",
            bf_jacobian,
            bf_jacobian,
            weights,
            det_jacs,
            optimize=True,
        )
        support = sp.utils.data.cartesian_product(
            (support[0, :], support[0, :])
        )
        system_matrix[support[:, 0], support[:, 1]] += local_matrix.flatten()

# Set dirichlet values (identify boundary nodes and set matrix to identity rhs
# to 0)
indices = np.hstack(
    (
        solution_field.multi_index[0, :],
        solution_field.multi_index[-1, :],
        solution_field.multi_index[:, 0],
        solution_field.multi_index[:, -1],
    )
)
system_matrix[indices, :] = 0
system_matrix[indices, indices] = 1
system_rhs[indices] = 0

# Solve linear system
solution_field.control_points = np.linalg.solve(
    system_matrix, system_rhs
).reshape(-1, 1)

# Plot geometry and field
geometry.spline_data["field"] = solution_field
geometry.show_options["data"] = "field"
geometry.show_options["cmap"] = "jet"
geometry.show_options["lighting"] = "off"
geometry.show_options["scalarbar"] = True
geometry.show(knots=True, control_points=False)



from functools import wraps as _wraps

import numpy as _np

from splinepy import settings as _settings


def valid_queries(spline, queries):
    """Check queries

    Check if queries are valid for a specific spline, so not to exceed
    parametric bounds for requests

    Parameters
    ----------
    spline : Spline-Type
      spline that sets the bounds
    queries : np.array
      Requested queries

    Returns
    -------
    valid : true
    """
    bounds = spline.parametric_bounds
    # Check dimensions
    if bounds.shape[1] != queries.shape[1]:
        raise ValueError(
            f"Dimension mismatch between parametric dimension of "
            f"spline ({spline.para_dim}), and query-request "
            f"({queries.shape[1]})."
        )

    # Check minimum value
    min_query = _np.min(queries, axis=0)
    if _np.any(bounds[0, :] > min_query):
        error_dim = _np.where(bounds[1, :] > min_query)[0][0]
        error_query = _np.argmin(queries, axis=0)[error_dim]
        raise ValueError(
            f"Query request out of bounds in parametric dimension "
            f"{error_dim}. Detected query {queries[error_query,:]} at "
            f"positions {error_query}, which is out of bounds with "
            f"minimum values {bounds[1,:]}."
        )

    # Check maximum value
    max_query = _np.max(queries, axis=0)
    if _np.any(bounds[1, :] < max_query):
        error_dim = _np.where(bounds[1, :] < max_query)[0][0]
        error_query = _np.argmax(queries, axis=0)[error_dim]
        raise ValueError(
            f"Query request out of bounds in parametric dimension "
            f"{error_dim}. Detected query {queries[error_query,:]} at "
            f"positions {error_query}, which is out of bounds with "
            f"maximum values {bounds[1,:]}."
        )
    return True


def clamped_knot_vectors(spline, warning=True):
    """
    Checks if knot vector is clamped. This also referred as open knot vector.
    If spline doesn't have enough information, this will return None.

    Parameters
    ---------
    spline: Spline

    Returns
    -------
    is_clamped: bool
    """
    if not spline.has_knot_vectors:
        return True

    degrees = spline.degrees
    knot_vectors = spline.knot_vectors
    if degrees is None or knot_vectors is None:
        return None

    for d, kv in zip(degrees, knot_vectors):
        kv_arr = _np.asanyarray(kv)

        front = all(abs(kv_arr[: (d + 1)] - kv_arr[0]) < _settings.TOLERANCE)
        end = all(abs(kv_arr[-(d + 1) :] - kv_arr[-1]) < _settings.TOLERANCE)

        if not front or not end:
            if warning:
                spline._logw(
                    "Spline has an unclamped/closed knot vector. "
                    "This is an experimental feature and may cause "
                    "unexpected behavior."
                )
            return False

    return True


class Checker:
    """Helper class to allow direct extraction from spline obj (BSpline or
    NURBS). Internal use only.

    Examples
    ---------

    .. code-block :: python

      spline.check.valid_queries(queries)

    Parameters
    ----------
    spline : Spline
      Parent spline, that is to be checked
    """

    __slots__ = ("_helpee",)

    def __init__(self, spl):
        self._helpee = spl

    @_wraps(valid_queries)
    def valid_queries(self, queries):
        return valid_queries(self._helpee, queries)

    @_wraps(clamped_knot_vectors)
    def clamped_knot_vectors(self, warning=True):
        return clamped_knot_vectors(self._helpee, warning)



import splinepy as spp

if __name__ == "__main__":
    # curve
    # define degrees
    ds1 = [2]

    # define knot vectors
    kvs1 = [[0.0, 0.0, 0.0, 1.0, 1.0, 1.0]]

    # define control points
    cps1 = [
        [0.0, 0.0],
        [-2.0, 4.0],
        [1.0, 3.0],
    ]

    # init bspline
    b = spp.BSpline(
        degrees=ds1,
        knot_vectors=kvs1,
        control_points=cps1,
    )

    # make richer
    b.elevate_degrees([0, 0])
    b.insert_knots(0, [0.1, 0.2, 0.4, 0.7, 0.8, 0.8, 0.8])

    # press `+` key to plot axes behind
    spp.show(
        ["BSpline", b],
        ["Basis", *b.extract.bases()],
        ["Basis (physical space)", *b.extract.bases(parametric_view=False)],
        resolutions=201,
    )

    # 2D Basis and 1D Basis
    # define degrees
    ds2 = [2, 2]

    # define knot vectors
    kvs2 = [
        [0.0, 0.0, 0.0, 0.5, 1.0, 1.0, 1.0],
        [0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
    ]

    # define control points
    cps2 = [
        [0, 0],
        [0, 1],
        [1, 1.5],
        [3, 1.5],
        [-1, 0],
        [-1, 2],
        [1, 4],
        [3, 4],
        [-2, 0],
        [-2, 2],
        [1, 5],
        [3, 5],
    ]

    # init bspline
    b = spp.BSpline(
        degrees=ds2,
        knot_vectors=kvs2,
        control_points=cps2,
    )

    b.elevate_degrees([0, 0, 1])
    b.insert_knots(0, [0.1, 0.6, 0.8, 0.8, 0.8])
    b.insert_knots(1, [0.4, 0.6, 0.8, 0.8, 0.8])

    spp.show(
        ["BSpline", b],
        ["Basis", *b.extract.bases()],
        ["Basis 11, 19, 43", *b.extract.bases([11, 19, 43])],
        ["Basis (physical space)", *b.extract.bases(parametric_view=False)],
        resolutions=101,
        knots=False,
    )
